# -----------------------------------------------------------------------------
# Klotho/klotho/chronos/rhythm_pairs/rp.py
# -----------------------------------------------------------------------------
'''
--------------------------------------------------------------------------------------
Rhythm pairs.
--------------------------------------------------------------------------------------
'''
from typing import Tuple
from math import prod
from ..rhythm_trees.algorithms import rhythm_pair

class RhythmPair:
    def __init__(self, lst: Tuple[int, ...], subdivs: bool = False):
        self.lst = lst
        self._subdivs = subdivs
        self._total_product = prod(lst)
        self._mm_sequence = rhythm_pair(self.lst, MM=True)
        self._non_mm_sequence = rhythm_pair(self.lst, MM=False)
        self._partitions = self._calculate_partitions()
        self._measures = self._calculate_measures()
        self._beats = self._non_mm_sequence

    @property
    def product(self) -> int:
        '''Return the total product of the list.'''
        return self._total_product
    
    @property
    def products(self) -> Tuple[int, int, int]:
        '''Return the products of the total product divided by each element in the list.'''
        return tuple(self._total_product // x for x in self.lst)

    @property
    def partitions(self) -> Tuple[Tuple[int, ...], ...]:
        '''Return partitions or just the partition labels depending on subdivs.'''
        if self._subdivs:
            return self._partitions
        return tuple(tuple(part[0] for part in group) for group in self._partitions)

    @property
    def measures(self) -> Tuple[int, ...]:
        '''Return measures or just the measure labels depending on subdivs.'''
        if self._subdivs:
            return self._measures
        return tuple(measure[0] for measure in self._measures)

    @property
    def beats(self) -> Tuple[int, ...]:
        '''Return the non-MM sequence (beats).'''
        return self._beats
    
    @property
    def subdivs(self) -> bool:
        '''Get the current state of subdivs.'''
        return self._subdivs
    
    @subdivs.setter
    def subdivs(self, value: bool):
        '''Set the subdivs flag.'''
        self._subdivs = value

    def _calculate_partitions(self) -> Tuple[Tuple[int, Tuple[int, ...]], ...]:
        '''Calculate partitions based on the non-MM sequence.'''
        mm_partitions = self.products
        return tuple(self._partition_sequence(self._non_mm_sequence, partition) for partition in mm_partitions)

    def _partition_sequence(self, sequence: Tuple[int, ...], partition_value: int) -> Tuple[int, Tuple[int, ...]]:
        '''Partition the sequence based on a given partition value.'''
        partitions = []
        current_partition = []
        current_sum = 0

        for value in sequence:
            current_partition.append(value)
            current_sum += value

            if current_sum == partition_value:
                partitions.append((partition_value, tuple(current_partition)))
                current_partition = []
                current_sum = 0

        return tuple(partitions)

    def _calculate_measures(self) -> Tuple[Tuple[int, Tuple[int, ...]], ...]:
        '''Calculate measures based on the MM sequence.'''
        partitions = []
        current_partition = []
        mm_index = 0
        current_sum = 0

        for value in self._non_mm_sequence:
            current_partition.append(value)
            current_sum += value

            if current_sum == self._mm_sequence[mm_index]:
                partitions.append((self._mm_sequence[mm_index], tuple(current_partition)))
                current_partition = []
                current_sum = 0
                mm_index = (mm_index + 1) % len(self._mm_sequence)

        return tuple(partitions)



"""
Composition: A specialized module for working with compositional structures.

This module provides tools for creating and manipulating composite musical structures
that span multiple domains (temporal, tonal, parametric, etc.).
"""

from .compositional import CompositionalUnit

__all__ = [
    'CompositionalUnit',
] 


from .pitch import Pitch
from .pitch_collections import (
    PitchCollection, 
    EquaveCyclicCollection, 
    AddressedPitchCollection,
    IntervalType,
    IntervalList,
    _addressed_collection_cache
)

__all__ = [
    'Pitch',
    'PitchCollection',
    'EquaveCyclicCollection', 
    'AddressedPitchCollection',
    'IntervalType',
    'IntervalList',
    '_addressed_collection_cache'
] 


from klotho.dynatos.dynamics import Dynamic, DynamicRange
from klotho.tonos.pitch import Pitch
from klotho.utils.data_structures.dictionaries import SafeDict
from typing import List, Dict, TypeVar, Union

class Instrument():
    
    def __init__(self,
                 name          = 'default',
                 pfields       = None
        ):
        """
        Initialize an Instrument.
        
        Args:
            name (str): The name of the instrument
            pfields (dict or SafeDict): Parameter fields with default values
        """
        self._name = name
        
        if pfields is None:
            pfields = {}
        self._pfields = pfields if isinstance(pfields, SafeDict) else SafeDict(pfields)
    
    @property
    def name(self):
        return self._name
    
    @property
    def pfields(self):
        return self._pfields.copy()
    
    def keys(self):
        keys = ['synth_name']
        keys.extend(self._pfields.keys())
        return keys
    
    def __getitem__(self, key):
        if key == 'synth_name':
            return self._name
        return self._pfields[key]
    
    def __str__(self):
        return f"Instrument(name='{self._name}', pfields={dict(self._pfields)})"
    
    def __repr__(self):
        return self.__str__()


class SynthDefInstrument(Instrument):
    
    def __init__(self,
                 name          = 'default',
                 freq_range    = None,
                 dynamic_range = None,
                 env_type      = 'Sustained',
                 pfields       = {'amp': 0.1, 'freq': 440.0, 'pan': 0.0, 'gate': 1, 'out': 0}
        ):
        """
        Initialize a SynthDefInstrument.
        
        Args:
            name (str): The name of the instrument
            freq_range (tuple): A tuple of (min, max) frequency values or Pitch instances
            dynamic_range: A DynamicRange instance or a tuple of (min, max) dB values
            env_type (str): The envelope type for the instrument
            pfields (dict or SafeDict): Parameter fields with default values
        """
        if pfields is None:
            pfields = {'amp': 0.1, 'freq': 440.0, 'pan': 0.0, 'gate': 1, 'out': 0}
        
        super().__init__(name=name, pfields=pfields)
        
        if freq_range is None:
            self._freq_range = (Pitch.from_freq(27.5), Pitch.from_freq(4186.01))
        else:
            self._freq_range = self._process_freq_range(freq_range)
        
        if dynamic_range is None:
            self._dynamic_range = DynamicRange(min_dynamic=-60, max_dynamic=-3, curve=1.25)
        else:
            self._dynamic_range = self._process_dynamic_range(dynamic_range)
        
        self._env_type = env_type
    
    def _process_freq_range(self, freq_range):
        min_freq, max_freq = freq_range
        
        if not isinstance(min_freq, Pitch):
            min_freq = Pitch.from_freq(float(min_freq))
        
        if not isinstance(max_freq, Pitch):
            max_freq = Pitch.from_freq(float(max_freq))
            
        return (min_freq, max_freq)
    
    def _process_dynamic_range(self, dynamic_range):
        if isinstance(dynamic_range, DynamicRange):
            return dynamic_range
        
        min_dyn, max_dyn = dynamic_range
        
        if isinstance(min_dyn, Dynamic):
            min_dyn = min_dyn.db
            
        if isinstance(max_dyn, Dynamic):
            max_dyn = max_dyn.db
            
        return DynamicRange(min_dynamic=min_dyn, max_dynamic=max_dyn)
    
    @property
    def freq_range(self):
        return self._freq_range
    
    @property
    def dynamic_range(self):
        return self._dynamic_range
    
    @property
    def env_type(self):
        return self._env_type
    
    def __str__(self):
        return f"SynthDefInstrument(name='{self._name}', pfields={dict(self._pfields)})"


class MidiInstrument(Instrument):
    
    def __init__(self,
                 name          = 'default',
                 prgm          = 0,
                 is_Drum       = False,
                 pfields       = None
        ):
        """
        Initialize a MidiInstrument.
        
        Args:
            name (str): The name of the instrument
            prgm (int): General MIDI program number (ignored if is_Drum is True)
            is_Drum (bool): Whether this instrument uses the general MIDI percussion channel
            pfields (dict or SafeDict): Parameter fields with default values
        """
        if pfields is None:
            pfields = {'note': 60 if not is_Drum else 35, 'velocity': 100}
        
        super().__init__(name=name, pfields=pfields)
        
        self._prgm = prgm
        self._is_Drum = is_Drum
    
    @property
    def prgm(self):
        return self._prgm
    
    @property
    def is_Drum(self):
        return self._is_Drum
    
    def __str__(self):
        return f"MidiInstrument(name='{self._name}', prgm={self._prgm}, is_Drum={self._is_Drum}, pfields={dict(self._pfields)})"



from klotho.utils.data_structures.enums import DirectValueEnumMeta, Enum
from collections import namedtuple
import numpy as np
from enum import member

A4_Hz   = 440.0
A4_MIDI = 69

__all__ = [
    'PITCH_CLASSES',
    'freq_to_midicents',
    'midicents_to_freq',
    'midicents_to_pitchclass',
    'freq_to_pitchclass',
    'pitchclass_to_freq',
    'A4_Hz',
    'A4_MIDI'
]

class PITCH_CLASSES(Enum, metaclass=DirectValueEnumMeta):
  @member
  class N_TET_12(Enum, metaclass=DirectValueEnumMeta):
    C  = 0
    Cs = 1
    Db = 1
    D  = 2
    Ds = 3
    Eb = 3
    E  = 4
    Es = 5
    Fb = 4
    F  = 5
    Fs = 6
    Gb = 6
    G  = 7
    Gs = 8
    Ab = 8
    A  = 9
    As = 10
    Bb = 10
    B  = 11
    Bs = 0
  
    @member
    class names:
      as_sharps = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
      as_flats  = ['C', 'Db', 'D', 'Eb', 'E', 'F', 'Gb', 'G', 'Ab', 'A', 'Bb', 'B']


def freq_to_midicents(frequency: float) -> float:
  '''
  Convert a frequency in Hertz to MIDI cents notation.
  
  MIDI cents are a logarithmic unit of measure used for musical intervals.
  The cent is equal to 1/100th of a semitone. There are 1200 cents in an octave.
  
  MIDI cents combines MIDI note numbers (denoting pitch with) with cents (denoting
  intervals).  The MIDI note number is the integer part of the value, and the cents
  are the fractional part.
  
  The MIDI note for A above middle C is 69, and the frequency is 440 Hz.  The MIDI
  cent value for A above middle C is 6900.  Adding or subtracting 100 to the MIDI
  cent value corresponds to a change of one semitone (one note number in the Western
  dodecaphonic equal-tempered "chromatic" scale).
  
  Values other than multiple of 100 indicate microtonal intervals.

  Args:
  frequency: The frequency in Hertz to convert.

  Returns:
  The MIDI cent value as a float.
  '''
  return 100 * (12 * np.log2(frequency / A4_Hz) + A4_MIDI)

def midicents_to_freq(midicents: float) -> float:
  '''
  Convert MIDI cents back to a frequency in Hertz.
  
  MIDI cents are a logarithmic unit of measure used for musical intervals.
  The cent is equal to 1/100th of a semitone. There are 1200 cents in an octave.
  
  MIDI cents combines MIDI note numbers (denoting pitch with) with cents (denoting
  intervals).  The MIDI note number is the integer part of the value, and the cents
  are the fractional part.
  
  The MIDI note for A above middle C is 69, and the frequency is 440 Hz.  The MIDI
  cent value for A above middle C is 6900.  Adding or subtracting 100 to the MIDI
  cent value corresponds to a change of one semitone (one note number in the Western
  dodecaphonic equal-tempered "chromatic" scale).
  
  Values other than multiple of 100 indicate microtonal intervals.
  
  Args:
    midicents: The MIDI cent value to convert.
    
  Returns:
    The corresponding frequency in Hertz as a float.
  '''
  return A4_Hz * (2 ** ((midicents - A4_MIDI * 100) / 1200.0))

def midicents_to_pitchclass(midicents: float) -> namedtuple:
  '''
  Convert MIDI cents to a pitch class with offset in cents.
  
  Args:
    midicents: The MIDI cent value to convert.
    
  Returns:
    A tuple containing the pitch class and the cents offset.
  '''
  result = namedtuple('result', ['pitchclass', 'octave', 'cents_offset'])
  PITCH_LABELS = PITCH_CLASSES.N_TET_12.names.as_sharps
  midi = midicents / 100
  midi_round = round(midi)
  note_index = int(midi_round) % len(PITCH_LABELS)
  octave = int(midi_round // len(PITCH_LABELS)) - 1  # MIDI starts from C-1
  pitch_label = PITCH_LABELS[note_index]
  cents_diff = (midi - midi_round) * 100
  return result(pitch_label, octave, round(cents_diff, 4))
  # return Pitch(pitch_label, octave, round(cents_diff, 4))
  
def freq_to_pitchclass(freq: float, cent_round: int = 4) -> namedtuple:
    '''
    Converts a frequency to a pitch class with offset in cents.
    
    Args:
        freq: The frequency in Hertz to convert.
        cent_round: Number of decimal places to round cents to
    
    Returns:
        A tuple containing the pitch class and the cents offset.
    '''
    result = namedtuple('result', ['pitchclass', 'octave', 'cents_offset'])
    PITCH_LABELS = PITCH_CLASSES.N_TET_12.names.as_sharps
    n_PITCH_LABELS = len(PITCH_LABELS)
    midi = A4_MIDI + n_PITCH_LABELS * np.log2(freq / A4_Hz)
    midi_round = round(midi)
    note_index = int(midi_round) % n_PITCH_LABELS
    octave = int(midi_round // n_PITCH_LABELS) - 1  # MIDI starts from C-1
    pitch_label = PITCH_LABELS[note_index]
    cents_diff = (midi - midi_round) * 100
    
    return result(pitch_label, octave, round(cents_diff, cent_round))

def pitchclass_to_freq(pitchclass: str, octave: int = 4, cent_offset: float = 0.0, hz_round: int = 4, A4_Hz=A4_Hz, A4_MIDI=A4_MIDI):
    '''
    Converts a pitch class with offset in cents to a frequency.
    
    Args:
        pitchclass: The pitch class (like "C4" or "F#-2") to convert.
        cent_offset: The cents offset, default is 0.0.
        A4_Hz: The frequency of A4, default is 440 Hz.
        A4_MIDI: The MIDI note number of A4, default is 69.
    
    Returns:
        The frequency in Hertz.
    '''
    # Try both sharp and flat notations
    SHARP_LABELS = PITCH_CLASSES.N_TET_12.names.as_sharps
    FLAT_LABELS = PITCH_CLASSES.N_TET_12.names.as_flats
    
    try:
        note_index = SHARP_LABELS.index(pitchclass)
    except ValueError:
        try:
            note_index = FLAT_LABELS.index(pitchclass)
        except ValueError:
            raise ValueError(f"Invalid pitch class: {pitchclass}")

    midi = note_index + (octave + 1) * 12
    midi = midi - A4_MIDI
    midi = midi + cent_offset / 100
    frequency = A4_Hz * (2 ** (midi / 12))
    return round(frequency, hz_round)



from klotho.utils.algorithms.factors import to_factors
from typing import Union, List, Tuple, Dict, Set
from collections import namedtuple
from fractions import Fraction
import numpy as np
from sympy import Rational, root
import pandas as pd

A4_Hz   = 440.0
A4_MIDI = 69

from klotho.utils.data_structures.enums import DirectValueEnumMeta, Enum  

__all__ = [
    'equave_reduce',
    'reduce_interval',
    'reduce_interval_relative',
    'reduce_sequence_relative',
    'fold_interval',
    'reduce_freq'
]

def equave_reduce(interval:Union[int, float, Fraction, str], equave:Union[Fraction, int, str, float] = 2, n_equaves:int = 1) -> Union[int, float, Fraction]:
  '''
  Reduce an interval to within the span of a specified octave.
  
  Args:
    interval: The musical interval to be octave-reduced.
    equave: The span of the octave for reduction, default is 2.
    n_equaves: The number of equaves, default is 1.
    
  Returns:
    The equave-reduced interval as a float.
  '''
  interval = Fraction(interval)
  equave = Fraction(equave)
  while interval < 1:
    interval *= equave
  while interval >= equave**n_equaves:
    interval /= equave
  return interval

def reduce_interval(interval:Union[Fraction, int, float, str], equave:Union[Fraction, int, float, str] = 2, n_equaves:int = 1) -> Fraction: 
  '''
  Fold an interval to within a specified range.

  Args:
    interval: The interval to be wrapped.
    equave: The equave value, default is 2.
    n_equaves: The number of equaves, default is 1.

  Returns:
    The folded interval as a float.
  '''
  interval = Fraction(interval)
  equave = Fraction(equave)  
  while interval < 1/(equave**n_equaves):
    interval *= equave
  while interval >= (equave**n_equaves):
    interval /= equave
  return interval

def reduce_interval_relative(target: Union[Fraction, int, float, str], source: Union[Fraction, int, float, str], equave: Union[Fraction, int, float, str] = 2) -> Fraction:
    '''
    Fold a target interval to minimize its distance from a source interval through octave reduction.

    Args:
        target: The interval to be folded
        source: The reference interval to fold relative to
        equave: The equave value, default is 2 (octave)

    Returns:
        The folded interval as a Fraction that minimizes distance from source
    '''
    target = Fraction(target)
    source = Fraction(source)
    equave = Fraction(equave)
    
    while target < 1:
        target *= equave
    while source < 1:
        source *= equave
        
    best_target = target
    min_distance = abs(source - target)
    
    test_up = target
    test_down = target
    while True:
        test_up *= equave
        test_down /= equave
        
        up_dist = abs(source - test_up)
        down_dist = abs(source - test_down)
        
        if up_dist < min_distance:
            min_distance = up_dist
            best_target = test_up
        elif down_dist < min_distance:
            min_distance = down_dist
            best_target = test_down
        else:
            break
            
    return best_target

def reduce_sequence_relative(sequence: List[Union[Fraction, int, float, str]], equave: Union[Fraction, int, float, str] = 2) -> List[Fraction]:
    '''
    Fold a sequence of intervals where each interval is folded relative to its neighbors.
    The first and last intervals remain unchanged, serving as anchors.
    Intermediate intervals are folded to minimize octave displacement between adjacent pairs.

    Args:
        sequence: List of intervals to be folded
        equave: The equave value, default is 2 (octave)

    Returns:
        List of folded intervals as Fractions, preserving original start and end values
    '''
    if len(sequence) <= 2:
        return [Fraction(x) for x in sequence]
    
    result = [Fraction(x) for x in sequence]
    
    for i in range(1, len(sequence)-1):
        result[i] = reduce_interval_relative(result[i], result[i-1], equave)
    
    for i in range(len(sequence)-2, 0, -1):
        result[i] = reduce_interval_relative(result[i], result[i+1], equave)
    
    return result
  
def fold_interval(interval: Union[Fraction, int, float, str], lower_thresh: Union[Fraction, int, float, str], upper_thresh: Union[Fraction, int, float, str]) -> Fraction:
    '''
    Fold an interval by reflecting it relative to explicit threshold boundaries.
    If interval exceeds upper threshold, measure how far above it is and move that 
    same distance down FROM the upper threshold.
    If interval is below lower threshold, measure how far below it is and move that
    same distance up FROM the lower threshold.

    Args:
        interval: The interval to be folded
        lower_thresh: The lower threshold interval
        upper_thresh: The upper threshold interval

    Returns:
        The folded interval as a Fraction
    '''
    interval = Fraction(interval)
    lower_thresh = Fraction(lower_thresh)
    upper_thresh = Fraction(upper_thresh)
    
    if interval > upper_thresh:
        distance = interval / upper_thresh
        return upper_thresh / distance
    elif interval < lower_thresh:
        distance = lower_thresh / interval
        return lower_thresh * distance
    
    return interval

def reduce_freq(freq: float, lower: float = 27.5, upper: float = 4186, equave: Union[int, float, Fraction, str] = 2) -> float:
  '''
  Fold a frequency value to within a specified range.
  
  Args:
    freq: The frequency to be wrapped.
    lower: The lower bound of the range.
    upper: The upper bound of the range.
    
  Returns:
    The folded frequency as a float.
  '''
  equave = Fraction(equave)
  while freq < lower:
      freq *= equave
  while freq > upper:
      freq /= equave  
  return float(freq)


"""
Node Identity Mapping System for NetworkX to RustworkX Migration

This module provides a bidirectional mapping system between arbitrary Python objects
and integer indices required by RustworkX. It maintains object identity while
providing O(1) lookup performance in both directions.
"""

from typing import Any, Optional, Dict, List, Set
import weakref
from collections.abc import Hashable


class NodeIdentityMapper:
    """
    Bidirectional mapping system between arbitrary node objects and integer indices.
    
    This class maintains the relationship between user-provided node objects and
    RustworkX's integer indices, ensuring consistent mapping across all graph operations.
    
    Features:
    - O(1) lookup performance in both directions
    - Handles non-hashable objects using object identity
    - Automatic cleanup of removed nodes
    - Thread-safe operations
    - Memory-efficient using weak references where possible
    """
    
    def __init__(self):
        # Primary mappings for hashable objects
        self._obj_to_index: Dict[Any, int] = {}
        self._index_to_obj: List[Optional[Any]] = []
        
        # Secondary mapping for non-hashable objects using object id
        self._nonhashable_to_index: Dict[int, int] = {}
        self._nonhashable_refs: Dict[int, Any] = {}
        
        # Set of freed indices for reuse
        self._freed_indices: Set[int] = set()
        
        # Next available index
        self._next_index = 0
    
    def add_node(self, node_obj: Any) -> int:
        """
        Add a node object and return its index.
        
        If the node already exists, returns its existing index.
        Otherwise, creates a new mapping and returns the new index.
        
        Args:
            node_obj: The node object to add (can be any Python object)
            
        Returns:
            int: The integer index assigned to this node
            
        Raises:
            ValueError: If node_obj is None (reserved value)
        """
        if node_obj is None:
            raise ValueError("None is not allowed as a node object")
        
        # Check if already mapped (hashable objects)
        if self._is_hashable(node_obj):
            if node_obj in self._obj_to_index:
                return self._obj_to_index[node_obj]
        else:
            # Check non-hashable objects by id
            obj_id = id(node_obj)
            if obj_id in self._nonhashable_to_index:
                return self._nonhashable_to_index[obj_id]
        
        # Get next available index
        if self._freed_indices:
            index = self._freed_indices.pop()
        else:
            index = self._next_index
            self._next_index += 1
        
        # Extend index list if necessary
        while len(self._index_to_obj) <= index:
            self._index_to_obj.append(None)
        
        # Store the mapping
        self._index_to_obj[index] = node_obj
        
        if self._is_hashable(node_obj):
            self._obj_to_index[node_obj] = index
        else:
            obj_id = id(node_obj)
            self._nonhashable_to_index[obj_id] = index
            self._nonhashable_refs[obj_id] = node_obj
        
        return index
    
    def get_index(self, node_obj: Any) -> Optional[int]:
        """
        Get the index for an existing node object.
        
        Args:
            node_obj: The node object to look up
            
        Returns:
            int: The index if found, None otherwise
        """
        if node_obj is None:
            return None
            
        if self._is_hashable(node_obj):
            return self._obj_to_index.get(node_obj)
        else:
            obj_id = id(node_obj)
            return self._nonhashable_to_index.get(obj_id)
    
    def get_object(self, index: int) -> Optional[Any]:
        """
        Get the original object from its index.
        
        Args:
            index: The integer index to look up
            
        Returns:
            Any: The original object if found, None otherwise
        """
        if 0 <= index < len(self._index_to_obj):
            return self._index_to_obj[index]
        return None
    
    def remove_node(self, node_obj: Any) -> bool:
        """
        Remove a node object and its mapping.
        
        Args:
            node_obj: The node object to remove
            
        Returns:
            bool: True if the node was found and removed, False otherwise
        """
        index = self.get_index(node_obj)
        if index is None:
            return False
        
        return self.remove_by_index(index)
    
    def remove_by_index(self, index: int) -> bool:
        """
        Remove a node by its index.
        
        Args:
            index: The integer index to remove
            
        Returns:
            bool: True if the index was found and removed, False otherwise
        """
        if not (0 <= index < len(self._index_to_obj)) or self._index_to_obj[index] is None:
            return False
        
        node_obj = self._index_to_obj[index]
        
        # Remove from primary mappings
        if self._is_hashable(node_obj):
            self._obj_to_index.pop(node_obj, None)
        else:
            obj_id = id(node_obj)
            self._nonhashable_to_index.pop(obj_id, None)
            self._nonhashable_refs.pop(obj_id, None)
        
        # Clear the index slot and mark as freed
        self._index_to_obj[index] = None
        self._freed_indices.add(index)
        
        return True
    
    def has_node(self, node_obj: Any) -> bool:
        """
        Check if a node object is mapped.
        
        Args:
            node_obj: The node object to check
            
        Returns:
            bool: True if the node exists in the mapping
        """
        return self.get_index(node_obj) is not None
    
    def has_index(self, index: int) -> bool:
        """
        Check if an index is mapped to a node.
        
        Args:
            index: The index to check
            
        Returns:
            bool: True if the index exists and is mapped
        """
        return (0 <= index < len(self._index_to_obj) and 
                self._index_to_obj[index] is not None)
    
    def clear(self):
        """Clear all mappings."""
        self._obj_to_index.clear()
        self._index_to_obj.clear()
        self._nonhashable_to_index.clear()
        self._nonhashable_refs.clear()
        self._freed_indices.clear()
        self._next_index = 0
    
    def get_all_objects(self) -> List[Any]:
        """
        Get all mapped node objects.
        
        Returns:
            List[Any]: List of all node objects currently mapped
        """
        return [obj for obj in self._index_to_obj if obj is not None]
    
    def get_all_indices(self) -> List[int]:
        """
        Get all mapped indices.
        
        Returns:
            List[int]: List of all indices currently mapped
        """
        return [i for i, obj in enumerate(self._index_to_obj) if obj is not None]
    
    def num_nodes(self) -> int:
        """
        Get the number of mapped nodes.
        
        Returns:
            int: Number of nodes currently mapped
        """
        return len(self._index_to_obj) - len(self._freed_indices)
    
    def _is_hashable(self, obj: Any) -> bool:
        """
        Check if an object is hashable.
        
        Args:
            obj: The object to check
            
        Returns:
            bool: True if the object is hashable
        """
        try:
            hash(obj)
            return True
        except TypeError:
            return False
    
    def __len__(self) -> int:
        """Return the number of mapped nodes."""
        return self.num_nodes()
    
    def __contains__(self, node_obj: Any) -> bool:
        """Check if a node object is mapped (supports 'in' operator)."""
        return self.has_node(node_obj)
    
    def __repr__(self) -> str:
        """String representation of the mapper."""
        return f"NodeIdentityMapper(nodes={self.num_nodes()}, next_index={self._next_index})" 


"""
Curve generation and mapping utilities for envelopes.

This module provides functions for generating various types of curves
including linear and exponential lines, arch/swell shapes, and value
mapping with curve transformations.
"""

import numpy as np

__all__ = [
    'line',
    'arch',
    'map_curve',
]

def line(start=0.0, end=1.0, steps=100, curve=0.0):
    '''
    Generate a curved line from start to end value over n steps.
    
    Args:
        start: Starting value
        end: Ending value
        steps: Number of steps
        curve: Shape of the curve. Negative for exponential, positive for logarithmic, 0 for linear
        
    Returns:
        numpy.ndarray: Array of values following the specified curve
    '''
    if curve == 0:
        return np.linspace(start, end, steps)
    
    t = np.linspace(0, 1, steps)
    curved_t = np.exp(curve * t) - 1
    curved_t = curved_t / (np.exp(curve) - 1)
    
    return start + (end - start) * curved_t

def arch(base=0.0, peak=1.0, steps=100, curve=0.0, axis=0):
    '''
    Generate a swelling curve that rises and falls, starting and ending at base value, peaking at peak value.
    
    Args:
        base: Starting and ending value
        peak: Peak value
        steps: Number of steps
        curve: Shape of the curve. Can be:
               - A single number: Same curve applied to both sides (negative for exponential, positive for logarithmic)
               - A tuple/list of two values: First value for ascending curve, second for descending
        axis: Position of the peak (-1 to 1). 0 centers the peak, negative shifts earlier, positive shifts later
        
    Returns:
        numpy.ndarray: Array of values following a swell curve
    '''
    axis = np.clip(axis, -1, 1)
    split_point = int((0.5 + axis * 0.4) * steps)
    
    if isinstance(curve, (list, tuple)) and len(curve) == 2:
        up_curve, down_curve = curve
    else:
        up_curve = down_curve = curve
    
    up = line(base, peak, split_point + 1, up_curve)
    down = line(peak, base, steps - split_point, down_curve)
    
    return np.concatenate([up[:-1], down])

def map_curve(value, in_range, out_range, curve=0.0):
    '''
    Map a value from an input range to an output range with optional curve shaping.
    
    Args:
        value: Input value to map
        in_range: Tuple of (min, max) for input range
        out_range: Tuple of (min, max) for output range
        curve: Shape of the curve. Negative for exponential, positive for logarithmic, 0 for linear
        
    Returns:
        float: Mapped value with curve applied
    '''
    normalized = np.interp(value, in_range, (0, 1))
    
    if curve != 0:
        normalized = np.exp(curve * normalized) - 1
        normalized = normalized / (np.exp(curve) - 1)
    
    return np.interp(normalized, (0, 1), out_range) 


import pytest
import numpy as np
from fractions import Fraction
from klotho.tonos.pitch import Pitch, PitchCollection, EquaveCyclicCollection, AddressedPitchCollection
from klotho.tonos.scales import Scale
from klotho.tonos.chords import Chord


class TestPitch:
    """Test the Pitch class functionality"""
    
    def test_pitch_default_construction(self):
        """Test pitch creation with default values"""
        p = Pitch()
        assert p.pitchclass == 'A'
        assert p.octave == 4
        assert p.cents_offset == 0.0
        assert p.partial == 1
        assert abs(p.freq - 440.0) < 1e-6
    
    def test_pitch_string_construction(self):
        """Test pitch creation from string notation"""
        # Basic pitch class with octave
        p1 = Pitch("C4")
        assert p1.pitchclass == "C"
        assert p1.octave == 4
        assert abs(p1.freq - 261.6256) < 1e-4
        
        # Sharp notation
        p2 = Pitch("F#5")
        assert p2.pitchclass == "F#"
        assert p2.octave == 5
        
        # Flat notation
        p3 = Pitch("Bb3")
        assert p3.pitchclass == "Bb"
        assert p3.octave == 3
        
        # Negative octave
        p4 = Pitch("C-1")
        assert p4.pitchclass == "C"
        assert p4.octave == -1
    
    def test_pitch_from_freq(self):
        """Test pitch creation from frequency"""
        # A4 = 440 Hz
        p1 = Pitch.from_freq(440.0)
        assert p1.pitchclass == "A"
        assert p1.octave == 4
        assert abs(p1.cents_offset) < 1e-6
        
        # C4 ≈ 261.63 Hz
        p2 = Pitch.from_freq(261.6256)
        assert p2.pitchclass == "C"
        assert p2.octave == 4
    
    def test_pitch_comparison(self):
        """Test pitch comparison operators"""
        p1 = Pitch("C4")
        p2 = Pitch("C4")
        p3 = Pitch("D4")
        p4 = Pitch("C5")
        
        # Equality
        assert p1 == p2
        assert p1 != p3
        
        # Ordering
        assert p1 < p3  # C4 < D4
        assert p1 < p4  # C4 < C5
        assert p3 > p1  # D4 > C4
        assert p1 <= p2
        assert p1 >= p2


class TestPitchCollection:
    """Test the PitchCollection base class"""
    
    def test_pitch_collection_defaults(self):
        """Test default construction"""
        pc = PitchCollection()
        expected_defaults = [Fraction(1, 1), Fraction(9, 8), Fraction(5, 4), 
                           Fraction(4, 3), Fraction(3, 2), Fraction(5, 3), Fraction(15, 8)]
        assert pc.degrees == expected_defaults
        assert pc.equave == Fraction(2, 1)
        assert pc.interval_type == Fraction
    
    def test_pitch_collection_custom_degrees(self):
        """Test construction with custom degrees"""
        # Fraction input (ratios)
        pc1 = PitchCollection(["1/1", "5/4", "3/2"], interval_type="ratios")
        assert len(pc1.degrees) == 3
        assert pc1.degrees[0] == Fraction(1, 1)
        assert pc1.degrees[1] == Fraction(5, 4)
        assert pc1.degrees[2] == Fraction(3, 2)
        assert pc1.interval_type == Fraction
        
        # Float input (cents)
        pc2 = PitchCollection([0.0, 386.3, 702.0], interval_type="cents")
        assert pc2.interval_type == float
        assert len(pc2.degrees) == 3
        assert pc2.degrees[0] == 0.0
        assert abs(pc2.degrees[1] - 386.3) < 1e-6
        assert abs(pc2.degrees[2] - 702.0) < 1e-6
        assert pc2.equave == 1200.0
    
    def test_pitch_collection_intervals(self):
        """Test interval calculation between degrees"""
        # Fraction intervals
        pc1 = PitchCollection(["1/1", "5/4", "3/2"], interval_type="ratios")
        intervals = pc1.intervals
        assert len(intervals) == 2
        assert intervals[0] == Fraction(5, 4)  # 5/4 / 1/1
        assert intervals[1] == Fraction(6, 5)  # (3/2) / (5/4)
        
        # Cents intervals
        pc2 = PitchCollection([0.0, 386.3, 702.0], interval_type="cents")
        intervals2 = pc2.intervals
        assert len(intervals2) == 2
        assert abs(intervals2[0] - 386.3) < 1e-6
        assert abs(intervals2[1] - 315.7) < 1e-6
    
    def test_pitch_collection_indexing(self):
        """Test various indexing methods"""
        pc = PitchCollection(["1/1", "5/4", "3/2", "2/1"])
        
        # Basic indexing
        assert pc[0] == Fraction(1, 1)
        assert pc[1] == Fraction(5, 4)
        assert pc[-1] == Fraction(2, 1)
        
        # Slice indexing
        subset = pc[1:3]
        assert isinstance(subset, PitchCollection)
        assert len(subset) == 2
        assert subset[0] == Fraction(5, 4)
        
        # Sequence indexing
        selected = pc[[0, 2]]
        assert isinstance(selected, PitchCollection)
        assert len(selected) == 2
        assert selected[0] == Fraction(1, 1)
        assert selected[1] == Fraction(3, 2)
    
    def test_pitch_collection_root_addressing(self):
        """Test creating addressed pitch collections"""
        pc = PitchCollection(["1/1", "5/4", "3/2"])
        
        # Root with string
        addressed = pc.root("C4")
        assert isinstance(addressed, AddressedPitchCollection)
        assert addressed.reference_pitch.pitchclass == "C"
        assert addressed.reference_pitch.octave == 4


class TestEquaveCyclicCollection:
    """Test the EquaveCyclicCollection class"""
    
    def test_equave_cyclic_defaults(self):
        """Test default construction with sorting and deduplication"""
        ecc = EquaveCyclicCollection()
        # Should be sorted and deduplicated version of defaults
        expected = [Fraction(1, 1), Fraction(9, 8), Fraction(5, 4), 
                   Fraction(4, 3), Fraction(3, 2), Fraction(5, 3), Fraction(15, 8)]
        assert ecc.degrees == expected
        assert ecc.equave == Fraction(2, 1)
    
    def test_equave_cyclic_sorting(self):
        """Test automatic sorting of degrees"""
        ecc = EquaveCyclicCollection(["3/2", "1/1", "5/4"])
        assert ecc.degrees == [Fraction(1, 1), Fraction(5, 4), Fraction(3, 2)]
    
    def test_equave_cyclic_deduplication(self):
        """Test automatic removal of duplicates"""
        ecc = EquaveCyclicCollection(["1/1", "5/4", "5/4", "3/2"])
        assert len(ecc.degrees) == 3
        assert ecc.degrees == [Fraction(1, 1), Fraction(5, 4), Fraction(3, 2)]
    
    def test_equave_cyclic_basic_indexing(self):
        """Test basic indexing without complex equave displacement"""
        ecc = EquaveCyclicCollection(["1/1", "5/4", "3/2"])
        
        # Basic indexing
        assert ecc[0] == Fraction(1, 1)
        assert ecc[1] == Fraction(5, 4)
        assert ecc[2] == Fraction(3, 2)
        
        # Simple next octave
        assert ecc[3] == Fraction(2, 1)  # 1/1 * 2/1
        
        # Simple previous octave
        assert ecc[-1] == Fraction(3, 4)  # 3/2 / 2/1
    
    def test_equave_cyclic_simple_slicing(self):
        """Test basic slice indexing"""
        ecc = EquaveCyclicCollection(["1/1", "5/4", "3/2"])
        
        # Basic slice within bounds
        subset1 = ecc[0:2]
        assert isinstance(subset1, PitchCollection)
        assert len(subset1) == 2
    
    def test_equave_cyclic_cents_mode(self):
        """Test EquaveCyclicCollection with cents"""
        # Basic cents collection
        ecc_cents = EquaveCyclicCollection([0.0, 386.3, 702.0], interval_type="cents")
        assert ecc_cents.interval_type == float
        assert len(ecc_cents.degrees) == 3
        assert ecc_cents.degrees[0] == 0.0
        assert abs(ecc_cents.degrees[1] - 386.3) < 1e-6
        assert ecc_cents.equave == 1200.0
        
        # Test infinite indexing with cents
        assert abs(ecc_cents[3] - 1200.0) < 1e-6  # 0.0 + 1200
        assert abs(ecc_cents[4] - 1586.3) < 1e-6  # 386.3 + 1200
        
        # Test negative indexing with cents
        assert abs(ecc_cents[-1] - (-498.0)) < 1e-6  # 702.0 - 1200
    
    def test_equave_cyclic_cents_deduplication(self):
        """Test deduplication with cents"""
        # Test that values within 1e-6 tolerance are deduplicated
        ecc = EquaveCyclicCollection([0.0, 386.3, 386.300001, 702.0], interval_type="cents")
        assert len(ecc.degrees) == 3  # Should deduplicate the close values


class TestAddressedPitchCollection:
    """Test the AddressedPitchCollection class"""
    
    def test_addressed_basic_functionality(self):
        """Test basic addressed pitch collection operations"""
        pc = PitchCollection(["1/1", "5/4", "3/2"])
        addressed = pc.root("C4")
        
        assert addressed.reference_pitch.pitchclass == "C"
        assert addressed.reference_pitch.octave == 4
        assert len(addressed) == 3
        
        # Check that we get Pitch objects back
        p0 = addressed[0]
        assert isinstance(p0, Pitch)
        assert p0.pitchclass == "C"
        assert p0.octave == 4
    
    def test_addressed_iteration(self):
        """Test iteration over addressed collection"""
        pc = PitchCollection(["1/1", "5/4", "3/2"])
        addressed = pc.root("C4")
        
        pitches = list(addressed)
        assert len(pitches) == 3
        assert all(isinstance(p, Pitch) for p in pitches)
        assert pitches[0].pitchclass == "C"


class TestScale:
    """Test the Scale class"""
    
    def test_scale_defaults(self):
        """Test default major scale construction"""
        scale = Scale()
        expected = [Fraction(1, 1), Fraction(9, 8), Fraction(5, 4), 
                   Fraction(4, 3), Fraction(3, 2), Fraction(5, 3), Fraction(15, 8)]
        assert scale.degrees == expected
        assert scale.equave == Fraction(2, 1)
    
    def test_scale_custom_construction(self):
        """Test construction with custom degrees"""
        # Pentatonic scale (ratios)
        penta = Scale(["1/1", "9/8", "5/4", "3/2", "5/3"], interval_type="ratios")
        assert len(penta.degrees) == 5
        assert penta.degrees[0] == Fraction(1, 1)
        assert penta.degrees[-1] == Fraction(5, 3)
        
        # Chromatic scale (cents) - now works!
        chromatic = Scale([i * 100.0 for i in range(12)], interval_type="cents")
        assert chromatic.interval_type == float
        assert len(chromatic.degrees) == 12
        assert chromatic.degrees[0] == 0.0
        assert chromatic.degrees[1] == 100.0
        assert chromatic.equave == 1200.0
    
    def test_scale_unison_insertion(self):
        """Test that unison is automatically inserted if missing"""
        # Without unison
        scale1 = Scale(["9/8", "5/4", "3/2"])
        assert scale1.degrees[0] == Fraction(1, 1)  # Auto-inserted
        assert len(scale1.degrees) == 4
        
        # With unison already present
        scale2 = Scale(["1/1", "9/8", "5/4"])
        assert scale2.degrees[0] == Fraction(1, 1)
        assert len(scale2.degrees) == 3  # No duplicate
    
    def test_scale_equave_removal(self):
        """Test that equave interval is removed from degrees"""
        # Scale with equave included - it should be removed
        scale = Scale(["1/1", "5/4", "3/2", "2/1"])
        # The 2/1 should be removed because it equals the equave
        assert Fraction(2, 1) not in scale.degrees
        assert len(scale.degrees) == 3
    
    def test_scale_basic_indexing(self):
        """Test basic indexing without complex operations"""
        scale = Scale(["1/1", "9/8", "5/4"])
        
        # Current octave
        assert scale[0] == Fraction(1, 1)
        assert scale[1] == Fraction(9, 8)
        assert scale[2] == Fraction(5, 4)
        
        # Next octave - simple cases
        assert scale[3] == Fraction(2, 1)    # 1/1 * 2/1
    
    def test_scale_mode_generation(self):
        """Test modal generation"""
        major = Scale(["1/1", "9/8", "5/4", "4/3", "3/2", "5/3", "15/8"])
        
        # Mode 0 should be the same scale
        mode0 = major.mode(0)
        assert mode0.degrees == major.degrees
        
        # Mode 1 (Dorian)
        dorian = major.mode(1)
        assert isinstance(dorian, Scale)
        assert dorian.degrees[0] == Fraction(1, 1)  # Always starts with unison
        assert len(dorian.degrees) == len(major.degrees)
    
    def test_scale_addressed_creation(self):
        """Test creating addressed scales"""
        scale = Scale(["1/1", "9/8", "5/4"])
        
        # Create addressed scale
        c_major = scale.root("C4")
        assert hasattr(c_major, '_collection')
        
        # Test indexing returns Pitch objects
        c4 = c_major[0]
        assert isinstance(c4, Pitch)
        assert c4.pitchclass == "C"
        assert c4.octave == 4


class TestChord:
    """Test the Chord class"""
    
    def test_chord_defaults(self):
        """Test default major triad construction"""
        chord = Chord()
        expected = [Fraction(1, 1), Fraction(5, 4), Fraction(3, 2)]
        assert chord.degrees == expected
        assert chord.equave == Fraction(2, 1)
    
    def test_chord_custom_construction(self):
        """Test construction with custom degrees"""
        # Minor triad (ratios)
        minor = Chord(["1/1", "6/5", "3/2"], interval_type="ratios")
        assert len(minor.degrees) == 3
        assert minor.degrees[1] == Fraction(6, 5)
        
        # Seventh chord (ratios)
        dom7 = Chord(["1/1", "5/4", "3/2", "7/4"], interval_type="ratios")
        assert len(dom7.degrees) == 4
        assert dom7.degrees[-1] == Fraction(7, 4)
        
        # Cents-based chord - now works!
        cents_chord = Chord([0.0, 386.3, 702.0, 968.8], interval_type="cents")
        assert cents_chord.interval_type == float
        assert len(cents_chord.degrees) == 4
        assert cents_chord.degrees[0] == 0.0
        assert abs(cents_chord.degrees[1] - 386.3) < 1e-6
        assert cents_chord.equave == 1200.0
    
    def test_chord_equave_behavior(self):
        """Test that chords handle equave like scales (remove it)"""
        # Chord with equave included - it should be removed like in Scale
        chord = Chord(["1/1", "5/4", "3/2", "2/1"])
        # The 2/1 should be removed because it equals the equave
        assert Fraction(2, 1) not in chord.degrees
        assert len(chord.degrees) == 3
    
    def test_chord_basic_indexing(self):
        """Test basic indexing"""
        chord = Chord(["1/1", "5/4", "3/2"])
        
        # Basic indexing
        assert chord[0] == Fraction(1, 1)
        assert chord[1] == Fraction(5, 4)
        assert chord[2] == Fraction(3, 2)
        
        # Next octave
        assert chord[3] == Fraction(2, 1)    # 1/1 * 2/1
    
    def test_chord_inversion_operations(self):
        """Test chord inversion methods"""
        chord = Chord(["1/1", "5/4", "3/2"])
        
        # Test invert operator
        inverted = ~chord
        assert isinstance(inverted, Chord)
        assert len(inverted.degrees) == len(chord.degrees)
        
        # Test negative operator (should be same as invert)
        negated = -chord
        assert negated.degrees == inverted.degrees
    
    def test_chord_addressed_creation(self):
        """Test creating addressed chords"""
        chord = Chord(["1/1", "5/4", "3/2"])
        
        # Create addressed chord
        c_major = chord.root("C4")
        assert hasattr(c_major, '_collection')
        
        # Test indexing returns Pitch objects
        c4 = c_major[0]
        assert isinstance(c4, Pitch)
        assert c4.pitchclass == "C"
        assert c4.octave == 4


class TestBasicIntegration:
    """Test basic integration between classes"""
    
    def test_pitch_collection_set_operations_fractions(self):
        """Test set operations with fraction collections only"""
        pc1 = PitchCollection(["1/1", "5/4", "3/2"])
        pc2 = PitchCollection(["1/1", "4/3", "3/2"])
        
        # Union
        union = pc1 | pc2
        assert len(union) == 4  # 1/1, 5/4, 4/3, 3/2
        assert Fraction(1, 1) in union.degrees
        assert Fraction(5, 4) in union.degrees
        assert Fraction(4, 3) in union.degrees
        assert Fraction(3, 2) in union.degrees
        
        # Intersection
        intersection = pc1 & pc2
        assert len(intersection) == 2  # 1/1, 3/2
        assert Fraction(1, 1) in intersection.degrees
        assert Fraction(3, 2) in intersection.degrees
    
    def test_real_world_musical_scenarios(self):
        """Test realistic musical scenarios"""
        # Major scale in C
        c_major_scale = Scale().root("C4")
        c4 = c_major_scale[0]  # C4
        d4 = c_major_scale[1]  # D4
        e4 = c_major_scale[2]  # E4
        
        assert c4.pitchclass == "C"
        assert d4.pitchclass == "D"
        assert e4.pitchclass == "E"
        
        # Major triad in different keys
        major_triad = Chord()
        
        # C major
        c_maj = major_triad.root("C4")
        assert c_maj[0].pitchclass == "C"
        
        # F# major
        fs_maj = major_triad.root("F#3")
        assert fs_maj[0].pitchclass == "F#"
        assert fs_maj[0].octave == 3
    
    def test_caching_behavior(self):
        """Test that addressed creation uses caching"""
        chord = Chord(["1/1", "5/4", "3/2"])
        scale = Scale(["1/1", "9/8", "5/4"])
        root_pitch = Pitch("C4")
        
        # Test that multiple calls to root() return cached objects
        addressed_chord1 = chord.root(root_pitch)
        addressed_chord2 = chord.root(root_pitch)
        assert addressed_chord1 is addressed_chord2
        
        addressed_scale1 = scale.root(root_pitch)
        addressed_scale2 = scale.root(root_pitch)
        assert addressed_scale1 is addressed_scale2


class TestCentsAndMixedOperations:
    """Test cents-based operations and mixed type operations"""
    
    def test_cents_scales_now_work(self):
        """FIXED: Cents-based scales now work with explicit interval_type"""
        # These now work with explicit interval_type
        scale1 = Scale([0.0, 100.0, 200.0], interval_type="cents")
        assert len(scale1.degrees) == 3
        assert scale1.interval_type == float
        assert scale1.equave == 1200.0
        
        chromatic = Scale([i * 100.0 for i in range(12)], interval_type="cents")
        assert len(chromatic.degrees) == 12
        assert chromatic.interval_type == float
        
        # Test addressing with cents scale
        c_chromatic = chromatic.root("C4")
        assert isinstance(c_chromatic[0], Pitch)
    
    def test_cents_chords_now_work(self):
        """FIXED: Cents-based chords now work with explicit interval_type"""
        chord1 = Chord([0.0, 400.0, 700.0], interval_type="cents")
        assert len(chord1.degrees) == 3
        assert chord1.interval_type == float
        assert chord1.equave == 1200.0
        
        # Test addressing with cents chord
        c_major_cents = chord1.root("C4")
        assert isinstance(c_major_cents[0], Pitch)
    
    def test_mixed_type_operations_work(self):
        """Test operations between different interval types"""
        fraction_pc = PitchCollection(["1/1", "5/4", "3/2"], interval_type="ratios")
        float_pc = PitchCollection([0.0, 386.3, 702.0], interval_type="cents")
        
        # Union should work by converting types
        union = fraction_pc | float_pc
        assert isinstance(union, PitchCollection)
        
        # Other operations should also work
        intersection = fraction_pc & float_pc
        assert isinstance(intersection, PitchCollection)
    
    def test_from_intervals_with_interval_type(self):
        """Test from_intervals with explicit interval_type"""
        # From ratio intervals
        pc1 = PitchCollection.from_intervals(["9/8", "10/9"], interval_type="ratios")
        assert pc1.interval_type == Fraction
        assert len(pc1) == 3  # Should include starting 1/1
        
        # From cents intervals
        pc2 = PitchCollection.from_intervals([203.9, 182.4], interval_type="cents")
        assert pc2.interval_type == float
        assert len(pc2) == 3
        assert pc2.degrees[0] == 0.0



from typing import Callable, Union, List, Tuple, Optional
import numpy as np
import pandas as pd
from ..lattices import Lattice


class Field(Lattice):
    """
    A field is a lattice with a function evaluated at each coordinate.
    
    Fields inherit all lattice functionality while providing field-specific
    methods for function evaluation and field manipulation. The function is
    evaluated lazily as coordinates are materialized in the lattice.
    
    Parameters
    ----------
    dimensionality : int
        Number of dimensions.
    resolution : int or list of int
        Number of points along each dimension, or list of resolutions per dimension.
    function : callable
        Function to evaluate at each coordinate. Should accept an array of shape
        (n_points, dimensionality) and return an array of shape (n_points,).
    ranges : tuple or list of tuple, optional
        Spatial range for each dimension. If tuple, applies to all dimensions.
        If list, must match dimensionality. Defaults to (-1, 1) per dimension.
    bipolar : bool, optional
        If True, coordinates range from -resolution to +resolution. 
        If False, coordinates range from 0 to resolution (default is True).
    periodic : bool, optional
        Whether to use periodic boundary conditions (default is False).
    """
    
    def __init__(self, 
                 dimensionality: int = 2, 
                 resolution: Union[int, List[int]] = 10, 
                 function: Callable[[np.ndarray], np.ndarray] = None,
                 ranges: Optional[Union[Tuple[float, float], List[Tuple[float, float]]]] = None,
                 bipolar: bool = True,
                 periodic: bool = False):
        
        if function is None:
            function = lambda x: np.zeros(x.shape[0])
        
        self._function = function
        
        if ranges is None:
            self._ranges = [(-1.0, 1.0)] * dimensionality
        elif isinstance(ranges, tuple) and len(ranges) == 2:
            self._ranges = [ranges] * dimensionality
        else:
            if len(ranges) != dimensionality:
                raise ValueError(f"Ranges list length {len(ranges)} must match dimensionality {dimensionality}")
            self._ranges = ranges
        
        super().__init__(dimensionality, resolution, bipolar, periodic)
        
        if not self._is_lazy:
            self._evaluate_all_coordinates()
    
    @property
    def function(self) -> Callable[[np.ndarray], np.ndarray]:
        """The function evaluated at each coordinate."""
        return self._function
    
    @property
    def ranges(self) -> List[Tuple[float, float]]:
        """Spatial ranges for each dimension."""
        return self._ranges.copy()
    
    def _coordinate_to_spatial_point(self, coord: Tuple[int, ...]) -> np.ndarray:
        """Convert a single integer lattice coordinate to spatial point."""
        spatial_point = []
        for i, c in enumerate(coord):
            if self._bipolar:
                coord_range = 2 * self._resolution[i]
                spatial_val = self._ranges[i][0] + (c + self._resolution[i]) * (self._ranges[i][1] - self._ranges[i][0]) / coord_range
            else:
                coord_range = self._resolution[i]
                spatial_val = self._ranges[i][0] + c * (self._ranges[i][1] - self._ranges[i][0]) / coord_range
            spatial_point.append(spatial_val)
        
        return np.array(spatial_point)
    
    def _coordinates_to_spatial_points(self, coords: List[Tuple[int, ...]]) -> np.ndarray:
        """Convert integer lattice coordinates to spatial points."""
        spatial_points = []
        
        for coord in coords:
            spatial_point = self._coordinate_to_spatial_point(coord)
            spatial_points.append(spatial_point)
        
        return np.array(spatial_points) if spatial_points else np.empty((0, self._dimensionality))
    
    def _evaluate_coordinates(self, coords: List[Tuple[int, ...]]):
        """Evaluate function at given coordinates and store values."""
        if not coords:
            return
        
        spatial_points = self._coordinates_to_spatial_points(coords)
        values = self._function(spatial_points)
        
        if np.isscalar(values):
            values = np.array([values])
        elif values.ndim == 0:
            values = np.array([values])
        
        for coord, value in zip(coords, values):
            node_id = self._coord_to_node[coord]
            current_data = self._graph.get_node_data(node_id) or {}
            current_data['field_value'] = float(value)
            self._graph[node_id] = current_data
    
    def _evaluate_all_coordinates(self):
        """Evaluate function at all existing coordinates."""
        coords = list(self._coord_to_node.keys())
        self._evaluate_coordinates(coords)
    
    def _materialize_coord(self, coord):
        """Override to evaluate function when materializing coordinates."""
        if not self._is_lazy or coord in self._materialized_coords:
            return
        
        coords_before = set(self._materialized_coords)
        super()._materialize_coord(coord)
        coords_after = set(self._materialized_coords)
        
        new_coords = list(coords_after - coords_before)
        if new_coords:
            self._evaluate_coordinates(new_coords)
    
    def get_field_value(self, coord: Tuple[int, ...]) -> float:
        """
        Get the field value at a specific coordinate.
        
        Parameters
        ----------
        coord : tuple of int
            The lattice coordinate.
            
        Returns
        -------
        float
            The field value at the coordinate.
        """
        node_id = self._get_node_for_coord(coord)
        if node_id is None:
            raise KeyError(f"Coordinate {coord} not found in field")
        
        node_data = self._graph.get_node_data(node_id) or {}
        return node_data.get('field_value', 0.0)
    
    def set_field_value(self, coord: Tuple[int, ...], value: float):
        """
        Set the field value at a specific coordinate.
        
        Parameters
        ----------
        coord : tuple of int
            The lattice coordinate.
        value : float
            The value to set.
        """
        node_id = self._get_node_for_coord(coord)
        if node_id is None:
            raise KeyError(f"Coordinate {coord} not found in field")
        
        current_data = self._graph.get_node_data(node_id) or {}
        current_data['field_value'] = float(value)
        self._graph[node_id] = current_data
    
    def apply_function(self, function: Callable[[np.ndarray], np.ndarray]):
        """
        Apply a new function to all lattice points, updating their values.
        
        Parameters
        ----------
        function : callable
            Function to evaluate at each coordinate. Should accept an array of shape
            (n_points, dimensionality) and return an array of shape (n_points,).
        """
        self._function = function
        
        if self._is_lazy:
            coords = list(self._materialized_coords)
        else:
            coords = list(self._coord_to_node.keys())
        
        self._evaluate_coordinates(coords)
    
    def sample_field(self, points: np.ndarray) -> np.ndarray:
        """
        Sample the field at arbitrary spatial points using interpolation.
        
        Parameters
        ----------
        points : numpy.ndarray
            Array of spatial points to sample, shape (n_points, dimensionality).
            
        Returns
        -------
        numpy.ndarray
            Array of interpolated field values.
        """
        from scipy.interpolate import griddata
        
        if self._is_lazy:
            coords = list(self._materialized_coords)
        else:
            coords = list(self._coord_to_node.keys())
        
        if not coords:
            return np.zeros(points.shape[0])
        
        spatial_points = self._coordinates_to_spatial_points(coords)
        field_values = np.array([self.get_field_value(coord) for coord in coords])
        
        return griddata(spatial_points, field_values, points, method='linear', fill_value=0.0)
    
    def gradient(self, coord: Tuple[int, ...]) -> np.ndarray:
        """
        Compute the gradient at a lattice coordinate using finite differences.
        
        Parameters
        ----------
        coord : tuple of int
            The lattice coordinate to compute gradient at.
            
        Returns
        -------
        numpy.ndarray
            Gradient vector at the coordinate.
        """
        gradient = np.zeros(self._dimensionality)
        neighbors = list(self.neighbors(coord))
        center_value = self.get_field_value(coord)
        
        for neighbor_coord in neighbors:
            direction = np.array(neighbor_coord) - np.array(coord)
            distance = np.linalg.norm(direction)
            if distance > 0:
                direction = direction / distance
                value_diff = self.get_field_value(neighbor_coord) - center_value
                gradient += direction * value_diff / distance
        
        return gradient
    
    def laplacian(self, coord: Tuple[int, ...]) -> float:
        """
        Compute the Laplacian at a lattice coordinate using finite differences.
        
        Parameters
        ----------
        coord : tuple of int
            The lattice coordinate to compute Laplacian at.
            
        Returns
        -------
        float
            Laplacian value at the coordinate.
        """
        neighbors = list(self.neighbors(coord))
        center_value = self.get_field_value(coord)
        neighbor_sum = sum(self.get_field_value(neighbor) for neighbor in neighbors)
        
        return neighbor_sum - len(neighbors) * center_value
    
    def get_field_values(self) -> List[float]:
        """
        Get all field values in the same order as coords.
        
        Returns
        -------
        list of float
            List of field values.
        """
        coords = self.coords
        return [self.get_field_value(coord) for coord in coords]
    
    def __getitem__(self, key):
        """Allow field[coordinate] access to values for tuples, otherwise delegate to parent."""
        if isinstance(key, tuple):
            return self.get_field_value(key)
        return super().__getitem__(key)
    
    def __setitem__(self, key, value):
        """Allow field[coordinate] = value setting for tuples."""
        if isinstance(key, tuple):
            self.set_field_value(key, value)
        else:
            raise TypeError("Only tuple keys are supported for field coordinate access")
    
    @classmethod
    def from_lattice(cls, lattice: Lattice, 
                     function: Callable[[np.ndarray], np.ndarray],
                     ranges: Optional[Union[Tuple[float, float], List[Tuple[float, float]]]] = None) -> 'Field':
        """
        Create a Field from an existing Lattice.
        
        Parameters
        ----------
        lattice : Lattice
            The lattice to convert to a field.
        function : callable
            Function to evaluate at each lattice point.
        ranges : tuple or list of tuple, optional
            Spatial ranges for the field.
            
        Returns
        -------
        Field
            A new Field instance with the same structure.
        """
        field = cls.__new__(cls)
        
        field._dimensionality = lattice._dimensionality
        field._resolution = lattice._resolution.copy()
        field._bipolar = lattice._bipolar
        field._periodic = lattice._periodic
        field._dims = lattice._dims.copy()
        field._coord_to_node = lattice._coord_to_node.copy()
        field._node_to_coord = lattice._node_to_coord.copy()
        field._materialized_coords = lattice._materialized_coords.copy()
        field._estimated_size = lattice._estimated_size
        field._is_lazy = lattice._is_lazy
        field._function = function
        
        field._graph = lattice._graph.copy()
        field._meta = lattice._meta.copy() if hasattr(lattice, '_meta') else pd.DataFrame(index=[''])
        
        if ranges is None:
            field._ranges = [(-1.0, 1.0)] * field._dimensionality
        elif isinstance(ranges, tuple) and len(ranges) == 2:
            field._ranges = [ranges] * field._dimensionality
        else:
            if len(ranges) != field._dimensionality:
                raise ValueError(f"Ranges list length {len(ranges)} must match dimensionality {field._dimensionality}")
            field._ranges = ranges
        
        if field._is_lazy:
            coords = list(field._materialized_coords)
        else:
            coords = list(field._coord_to_node.keys())
        
        field._evaluate_coordinates(coords)
        
        return field
    
    def __str__(self) -> str:
        """String representation of the field."""
        if self._is_lazy:
            coord_count = f"{len(self._materialized_coords)} materialized"
        else:
            coord_count = str(len(self.coords))
        
        return (f"Field(dimensionality={self._dimensionality}, "
                f"resolution={self._resolution}, "
                f"bipolar={self._bipolar}, "
                f"coordinates={coord_count})")
    
    def __repr__(self) -> str:
        return self.__str__()


"""
Test script to verify benchmarking infrastructure functionality.

This script validates that the benchmarking framework correctly measures
timing performance with acceptable variance and produces consistent results.
"""
from klotho.benchmarks.core import BenchmarkRunner
from klotho.topos.graphs import Graph
import time


def test_basic_timing():
    """Test basic timing functionality with known operations."""
    runner = BenchmarkRunner(warmup_runs=2, timing_runs=5)
    
    # Test with a predictable sleep operation
    def sleep_10ms():
        time.sleep(0.01)  # 10ms sleep
    
    result = runner.benchmark(sleep_10ms, "sleep_10ms_test")
    
    # Verify timing is approximately correct (within reasonable variance)
    expected_time = 0.01  # 10ms
    assert 0.008 <= result.mean <= 0.015, f"Sleep timing off: {result.mean}"
    assert result.coefficient_of_variation() < 10, f"High variance: {result.coefficient_of_variation()}%"
    
    print(f"✓ Basic timing test passed: {result.mean*1000:.2f}ms ± {result.stdev*1000:.2f}ms")


def test_graph_creation_benchmark():
    """Test benchmarking with actual graph operations."""
    runner = BenchmarkRunner(warmup_runs=2, timing_runs=5)
    
    # Test empty graph creation
    def create_empty_graph():
        graph = Graph()
        return graph
    
    result = runner.benchmark(create_empty_graph, "empty_graph_creation")
    
    # Should be very fast (< 1ms typically)
    assert result.mean < 0.001, f"Graph creation too slow: {result.mean*1000:.2f}ms"
    
    print(f"✓ Graph creation test passed: {result.mean*1000:.3f}ms")


def test_node_addition_benchmark():
    """Test benchmarking node addition operations."""
    runner = BenchmarkRunner(warmup_runs=2, timing_runs=5)
    
    # Test adding 100 nodes
    def add_nodes():
        graph = Graph()
        for i in range(100):
            graph.add_node(value=i)
        return graph
    
    result = runner.benchmark(add_nodes, "add_100_nodes")
    
    # Should complete reasonably quickly
    assert result.mean < 0.01, f"Node addition too slow: {result.mean*1000:.2f}ms"
    
    print(f"✓ Node addition test passed: {result.mean*1000:.2f}ms for 100 nodes")


def test_consistency():
    """Test that repeated benchmarks give consistent results."""
    runner = BenchmarkRunner(warmup_runs=2, timing_runs=10)
    
    def simple_graph_op():
        graph = Graph()
        for i in range(10):
            graph.add_node(value=i)
        return len(graph)
    
    # Run the same benchmark twice
    result1 = runner.benchmark(simple_graph_op, "consistency_test_1")
    result2 = runner.benchmark(simple_graph_op, "consistency_test_2")
    
    # Results should be similar (within 20% of each other)
    ratio = max(result1.mean, result2.mean) / min(result1.mean, result2.mean)
    assert ratio < 1.2, f"Inconsistent results: {result1.mean} vs {result2.mean}"
    
    print(f"✓ Consistency test passed: {result1.mean*1000:.3f}ms vs {result2.mean*1000:.3f}ms")


def main():
    """Run all infrastructure tests."""
    print("Testing Klotho Benchmarking Infrastructure")
    print("=" * 50)
    
    try:
        test_basic_timing()
        test_graph_creation_benchmark()
        test_node_addition_benchmark()
        test_consistency()
        
        print("\n🎉 All infrastructure tests passed!")
        print("Benchmarking framework is ready for use.")
        
    except Exception as e:
        print(f"\n❌ Test failed: {e}")
        raise


if __name__ == "__main__":
    main() 


import rustworkx as rx
import copy
from functools import lru_cache
from typing import List, TypeVar, Optional, Any, Union, Dict, Iterator, Tuple
import numpy as np

T = TypeVar('T')


class Graph:
    def __init__(self, graph: Union[rx.PyGraph, rx.PyDiGraph, 'Graph', None] = None):
        """Initialize a Graph with optional existing graph data.
        
        Args:
            graph: Optional existing graph to copy from (RustworkX graph, Graph instance, or None)
        """
        self._meta = {}
        self._structure_version = 0
        
        if graph is None:
            self._graph = rx.PyGraph()
        elif isinstance(graph, Graph):
            self._graph = graph._graph.copy()
            self._meta = copy.deepcopy(graph._meta)
            self._structure_version = 0
        elif isinstance(graph, (rx.PyGraph, rx.PyDiGraph)):
            self._graph = graph.copy()
        else:
            raise TypeError(f"Unsupported graph type: {type(graph)}")   
        
    @property
    def nodes(self):
        """Return a view of the nodes that can be subscripted."""
        return GraphNodeView(self)
    
    @property
    def edges(self):
        """Return a view of the edges."""
        return GraphEdgeView(self)
    
    def __getitem__(self, node):
        """Get node data for a given node."""
        if not self._graph.has_node(node):
            raise KeyError(f"Node {node} not found in graph")
        node_data = self._graph.get_node_data(node)
        
        if not isinstance(node_data, dict):
            empty_dict = {}
            return empty_dict
        
        return node_data
    
    def __len__(self):
        """Return the number of nodes."""
        return self._graph.num_nodes()
    
    def __str__(self):
        """String representation of the graph."""
        return f"Graph with {self.number_of_nodes()} nodes and {self.number_of_edges()} edges"
    
    def __repr__(self):
        """String representation of the graph."""
        return f"Graph({self.number_of_nodes()}, {self.number_of_edges()})"
    
    def __iter__(self):
        """Iterate over node objects."""
        return iter(self._graph.node_indices())
    
    def __contains__(self, node):
        """Check if a node is in the graph."""
        return self._graph.has_node(node)
    
    def _invalidate_caches(self):
        """Invalidate all caches when structure changes"""
        self._structure_version += 1
        if hasattr(self, 'descendants'):
            self.descendants.cache_clear()
        if hasattr(self, 'ancestors'):
            self.ancestors.cache_clear()
        if hasattr(self, 'successors'):
            self.successors.cache_clear()
        if hasattr(self, 'predecessors'):
            self.predecessors.cache_clear()
    
    def out_degree(self, node):
        """Get the out-degree of a node"""
        if hasattr(self._graph, 'out_degree'):
            return self._graph.out_degree(node)
        else:
            return self._graph.degree(node)
    
    def in_degree(self, node):
        """Get the in-degree of a node"""
        if hasattr(self._graph, 'in_degree'):
            return self._graph.in_degree(node)
        else:
            return self._graph.degree(node)
    
    def _get_node_object(self, index):
        """Convert RustworkX node index to node object.
        
        For the base Graph class, nodes are just their indices.
        Subclasses can override this for different node representations.
        """
        return index
    
    def _get_node_index(self, node):
        """Convert node object to RustworkX index.
        
        For the base Graph class, nodes are just their indices.
        Subclasses can override this for different node representations.
        """
        return node
    
    @classmethod
    def digraph(cls):
        """Create a new directed graph instance."""
        graph = cls()
        graph._graph = rx.PyDiGraph()
        return graph
    
    def neighbors(self, node):
        """Get neighbors of a node"""
        return list(self._graph.neighbors(node))
    
    @lru_cache(maxsize=None)
    def predecessors(self, node):
        """Returns all predecessors of a node.
        
        Args:
            node: The node whose predecessors to return
            
        Returns:
            tuple: All predecessors of the node
        """
        _ = self._structure_version
        if hasattr(self._graph, 'predecessor_indices'):
            return tuple(self._graph.predecessor_indices(node))
        else:
            return tuple(self.neighbors(node))
    
    @lru_cache(maxsize=None)
    def successors(self, node):
        """Returns all successors of a node.
        
        Args:
            node: The node whose successors to return
            
        Returns:
            tuple: All successors of the node in sorted order (left-to-right)
        """
        _ = self._structure_version
        if hasattr(self._graph, 'successor_indices'):
            succ_indices = self._graph.successor_indices(node)
            return tuple(sorted(succ_indices))
        else:
            return tuple(sorted(self.neighbors(node)))
    
    @lru_cache(maxsize=None)
    def descendants(self, node):
        """Returns all descendants of a node using native RustworkX algorithm.
        
        Args:
            node: The node whose descendants to return
            
        Returns:
            tuple: All descendants of the node
        """
        _ = self._structure_version
        try:
            return tuple(rx.descendants(self._graph, node))
        except Exception:
            return tuple()
    
    @lru_cache(maxsize=None)
    def ancestors(self, node):
        """Returns all ancestors of a node using native RustworkX algorithm.
        
        Args:
            node: The node whose ancestors to return
            
        Returns:
            tuple: All ancestors of the node
        """
        _ = self._structure_version
        try:
            return tuple(rx.ancestors(self._graph, node))
        except Exception:
            return tuple()
    
    def topological_sort(self):
        """Returns nodes in topological order.
        
        Returns:
            generator: Nodes in topological order
        """
        if hasattr(self._graph, 'out_degree'):
            indices = rx.topological_sort(self._graph)
        else:
            indices = self._graph.node_indices()
        
        return (idx for idx in indices)
    
    def to_directed(self):
        """Return a directed version of this graph.
        
        Returns:
            Graph: A new Graph instance with directed edges
        """
        directed_rx = rx.PyDiGraph()
        
        for idx in self._graph.node_indices():
            node_data = self._graph.get_node_data(idx)
            directed_rx.add_node(node_data)
        
        for src, tgt, edge_data in self.edges(data=True):
            directed_rx.add_edge(src, tgt, edge_data)
        
        new_graph = Graph.__new__(Graph)
        new_graph._graph = directed_rx
        new_graph._meta = copy.deepcopy(self._meta)
        new_graph._structure_version = 0
        
        return new_graph
    
    def number_of_nodes(self):
        """Return the number of nodes in the graph.
        
        Returns:
            int: Number of nodes
        """
        return self._graph.num_nodes()
        
    def number_of_edges(self):
        """Return the number of edges in the graph.
        
        Returns:
            int: Number of edges
        """
        return self._graph.num_edges()
        
    def nodes_with_data(self, data=True):
        """Return nodes with their data.
        
        Args:
            data: If True, return node data as well
            
        Returns:
            Iterator: Iterator of (node, data) pairs if data=True, else just nodes
        """
        if data:
            for idx in self._graph.node_indices():
                node_data = self._graph.get_node_data(idx)
                yield (idx, node_data if isinstance(node_data, dict) else {})
        else:
            for idx in self._graph.node_indices():
                yield idx
    
    def subgraph(self, node, renumber=True):
        """Extract a subgraph starting from a given node.
        
        Args:
            node: The node to use as the starting point of the subgraph
            renumber: Whether to renumber the nodes in the new graph
            
        Returns:
            Graph: A new Graph object representing the subgraph
        """
        if node not in self:
            raise ValueError(f"Node {node} not found in graph")
            
        descendants = [node] + list(self.descendants(node))
        
        subgraph_rx = self._graph.subgraph(descendants)
        
        return self._from_graph(subgraph_rx, renumber=renumber)
    
    @property
    def root_nodes(self):
        """Returns root nodes (nodes with no predecessors)"""
        root_indices = []
        
        if hasattr(self._graph, 'in_degree'):
            for idx in self._graph.node_indices():
                if self._graph.in_degree(idx) == 0:
                    root_indices.append(idx)
        else:
            if self._graph.num_nodes() == 0:
                return tuple()
            
            degrees = [(idx, self._graph.degree(idx)) for idx in self._graph.node_indices()]
            if degrees:
                min_deg_nodes = [idx for idx, deg in degrees if deg > 0]
                if min_deg_nodes:
                    root_indices = [min(min_deg_nodes)]
        
        return tuple(root_indices)
    

        
    def add_node(self, **attr):
        """Add a node to the graph.
        
        Args:
            **attr: Node attributes
            
        Returns:
            The node ID that was added
        """
        node_id = self._graph.add_node(attr if attr else {})
        self._invalidate_caches()
        return node_id
    
    def set_node_data(self, node, **attr):
        """Update data for an existing node.
        
        Args:
            node: The node to update
            **attr: Node attributes to set
        """
        if not self._graph.has_node(node):
            raise KeyError(f"Node {node} not found in graph")
        
        existing_data = self._graph.get_node_data(node)
        if existing_data is None:
            new_data = attr.copy()
            node_data = {}
            node_data.update(attr)
            self._graph.remove_node(node)
            new_node_id = self._graph.add_node(attr)
            if new_node_id != node:
                raise RuntimeError(f"Expected node ID {node}, got {new_node_id}")
        elif not isinstance(existing_data, dict):
            raise ValueError(f"Cannot update non-dict node data for node {node}. Node data type: {type(existing_data)}")
        else:
            existing_data.update(attr)
        
        self._invalidate_caches()
    
    def remove_node(self, node):
        """Remove a node from the graph."""
        self._graph.remove_node(node)
        self._invalidate_caches()
        
    def add_edge(self, u, v, **attr):
        """Add an edge to the graph with optional attributes."""
        if not self._graph.has_node(u):
            raise KeyError(f"Node {u} not found in graph")
        if not self._graph.has_node(v):
            raise KeyError(f"Node {v} not found in graph")
        
        self._graph.add_edge(u, v, attr if attr else {})
        self._invalidate_caches()
        
    def has_edge(self, u, v):
        """Check if an edge exists between two nodes."""
        return self._graph.has_edge(u, v)
        
    def remove_edge(self, u, v):
        """Remove an edge from the graph."""
        self._graph.remove_edge(u, v)
        self._invalidate_caches()
        
    def update(self, edges=None, nodes=None):
        """Update the graph with nodes and edges."""
        if nodes:
            for node_data in nodes:
                if isinstance(node_data, tuple) and len(node_data) == 2:
                    node, attrs = node_data
                    self.add_node(node, **attrs)
                else:
                    self.add_node(node_data)
        
        if edges:
            for edge_data in edges:
                if len(edge_data) == 2:
                    u, v = edge_data
                    self.add_edge(u, v)
                elif len(edge_data) == 3:
                    u, v, attrs = edge_data
                    self.add_edge(u, v, **attrs)
        
        self._invalidate_caches()
        
    def clear(self):
        """Remove all nodes and edges from the graph."""
        self._graph.clear()
        self._invalidate_caches()
    
    def set_node_attributes(self, node, attributes):
        """Set attributes for a node."""
        current_data = self._graph.get_node_data(node)
        if not isinstance(current_data, dict):
            current_data = {}
        current_data.update(attributes)
    
    def clear_node_attributes(self, nodes=None):
        """Clear attributes of specified nodes or all nodes.
        
        Args:
            nodes: Specific nodes to clear attributes for, or None for all nodes
        """
        nodes_to_clear = nodes if nodes is not None else self._graph.node_indices()
        for node in nodes_to_clear:
            if node in self:
                self._graph[node] = {}
        
    def renumber_nodes(self, method='default'):
        """Renumber the nodes in the graph to consecutive integers.
        
        Args:
            method (str): The method to use for renumbering:
                - 'default': Use sequential numbering
                - 'dfs': Use depth-first search preorder
                - 'bfs': Use breadth-first search
                
        Returns:
            Graph: Self with renumbered nodes
        """
        if method == 'default':
            pass
        elif method in ['dfs', 'bfs']:
            pass
        else:
            raise ValueError(f"Unknown renumbering method: {method}")
            
        return self

    def copy(self):
        """Create a deep copy of this graph."""
        return copy.deepcopy(self)
    
    def is_directed(self):
        """Return True if graph is directed, False otherwise."""
        return isinstance(self._graph, rx.PyDiGraph)
    
    def is_multigraph(self):
        """Return True if graph is a multigraph, False otherwise."""
        return False
    
    def to_networkx(self):
        """Convert this Graph to a NetworkX graph for compatibility with NetworkX functions.
        
        Returns:
            networkx.Graph or networkx.DiGraph: NetworkX equivalent of this graph
        """
        import networkx as nx
        
        if self.is_directed():
            nx_graph = nx.DiGraph()
        else:
            nx_graph = nx.Graph()
        
        for node, attrs in self.nodes(data=True):
            nx_graph.add_node(node, **attrs)
        
        for u, v, attrs in self.edges(data=True):
            nx_graph.add_edge(u, v, **attrs)
        
        return nx_graph

    @classmethod
    def _from_graph(cls, G, **kwargs):
        """Create a new instance from an existing graph.
        
        Args:
            G: The graph to create a new instance from
            **kwargs: Additional arguments
            
        Returns:
            Graph: A new Graph instance
        """
        return cls(G)
    
    @classmethod
    def grid_graph(cls, dims, periodic=False):
        """Create a grid graph with dimensions specified in dims.
        
        Creates a grid graph where nodes are numbered sequentially 0, 1, 2, ...
        and node data contains the coordinate as a tuple.
        
        Parameters
        ----------
        dims : list or tuple
            List of ranges or iterables defining the coordinate space for each dimension.
        periodic : bool, optional
            Whether to create periodic boundary conditions (default is False)
            
        Returns
        -------
        Graph
            A new Graph instance with grid structure and coordinate data in nodes
        """
        if not dims or len(dims) == 0:
            return cls()
        
        dims = [list(d) for d in dims]
        
        if len(dims) == 1:
            if periodic:
                rx_graph = rx.generators.cycle_graph(len(dims[0]))
            else:
                rx_graph = rx.generators.path_graph(len(dims[0]))
            for i, coord_val in enumerate(dims[0]):
                rx_graph[i] = {'coord': (coord_val,)}
        elif len(dims) == 2 and not periodic:
            rows, cols = len(dims[0]), len(dims[1])
            rx_graph = rx.generators.grid_graph(rows, cols)
            import itertools
            for i, coord in enumerate(itertools.product(dims[0], dims[1])):
                rx_graph[i] = {'coord': coord}
        else:
            rx_graph = rx.PyGraph()
            
            import itertools
            all_coords = list(itertools.product(*dims))
            
            node_indices = rx_graph.add_nodes_from([{'coord': coord} for coord in all_coords])
            coord_to_index = {coord: i for i, coord in enumerate(all_coords)}
            
            edge_list = []
            for coord in all_coords:
                coord_idx = coord_to_index[coord]
                
                for dim_idx in range(len(coord)):
                    for direction in [-1, 1]:
                        neighbor_coord = list(coord)
                        neighbor_coord[dim_idx] += direction
                        neighbor_coord = tuple(neighbor_coord)
                        
                        if neighbor_coord in coord_to_index:
                            neighbor_idx = coord_to_index[neighbor_coord]
                            if coord_idx < neighbor_idx:
                                edge_list.append((coord_idx, neighbor_idx))
                        elif periodic:
                            dim_values = dims[dim_idx]
                            if direction == 1 and coord[dim_idx] == dim_values[-1]:
                                wrap_coord = list(coord)
                                wrap_coord[dim_idx] = dim_values[0]
                                wrap_coord = tuple(wrap_coord)
                                if wrap_coord in coord_to_index:
                                    wrap_idx = coord_to_index[wrap_coord]
                                    if coord_idx < wrap_idx:
                                        edge_list.append((coord_idx, wrap_idx))
                            elif direction == -1 and coord[dim_idx] == dim_values[0]:
                                wrap_coord = list(coord)
                                wrap_coord[dim_idx] = dim_values[-1]
                                wrap_coord = tuple(wrap_coord)
                                if wrap_coord in coord_to_index:
                                    wrap_idx = coord_to_index[wrap_coord]
                                    if coord_idx < wrap_idx:
                                        edge_list.append((coord_idx, wrap_idx))
            
            if edge_list:
                rx_graph.extend_from_edge_list(edge_list)
        
        return cls(rx_graph)
    
    @classmethod
    def complete_graph(cls, n_nodes):
        """Create a complete graph.
        
        Args:
            n_nodes: Number of nodes in the complete graph
            
        Returns:
            Graph: A new Graph instance with complete structure
        """
        rx_graph = rx.generators.complete_graph(n_nodes)
        graph = cls(rx_graph)
        
        for node in range(n_nodes):
            if not isinstance(graph._graph.get_node_data(node), dict):
                graph._graph.remove_node(node)
                graph._graph.add_node({})
                
        for i in range(n_nodes):
            for j in range(i + 1, n_nodes):
                graph._graph.add_edge(i, j, {})
        
        return graph
    
    @classmethod
    def digraph(cls):
        """Create a directed graph.
        
        Returns:
            Graph: A new Graph instance with directed structure
        """
        new_graph = cls.__new__(cls)
        new_graph._graph = rx.PyDiGraph()
        new_graph._meta = {}
        new_graph._structure_version = 0
        return new_graph
    
    @classmethod
    def from_cost_matrix(cls, cost_matrix: np.ndarray, items: List[T]):
        """Create a Graph from a cost matrix.
        
        Transform a symmetric cost matrix into an undirected graph where
        edge weights represent the costs between nodes. Self-loops are
        excluded from the resulting graph.

        Parameters
        ----------
        cost_matrix : numpy.ndarray
            Symmetric cost matrix with numeric values. Should be square
            with dimensions matching the length of items.
        items : List[T]
            List of items corresponding to matrix indices. Used as node
            values in the resulting graph.

        Returns
        -------
        Graph
            Undirected graph with nodes corresponding to matrix indices
            and edge weights equal to the cost matrix values. Only edges
            with positive costs are included. Node attributes 'value' 
            contain the original items.

        Examples
        --------
        Create a graph from a simple cost matrix:
        
        >>> import numpy as np
        >>> matrix = np.array([[0, 1, 2], [1, 0, 3], [2, 3, 0]])
        >>> items = ['A', 'B', 'C']
        >>> graph = Graph.from_cost_matrix(matrix, items)
        >>> list(graph.edges(data=True))
        [('A', 'B', {'weight': 1}), ('A', 'C', {'weight': 2}), ('B', 'C', {'weight': 3})]
        """
        graph = cls()
        
        node_list = []
        for item in items:
            node_id = graph.add_node(value=item)
            node_list.append((node_id, {'value': item}))
        
        edge_list = []
        for i in range(len(items)):
            for j in range(i + 1, len(items)):
                cost = cost_matrix[i, j]
                if cost > 0:
                    edge_list.append((node_list[i][0], node_list[j][0], {'weight': cost}))
        
        if edge_list:
            graph._graph.add_edges_from(edge_list)
        
        return graph
    
    def __deepcopy__(self, memo):
        new_graph = self.__class__.__new__(self.__class__)
        
        new_graph._graph = self._graph.copy()
        new_graph._meta = copy.deepcopy(self._meta, memo)
        new_graph._structure_version = 0
        
        return new_graph


class GraphNodeView:
    """View of graph nodes that mimics NetworkX NodeView behavior."""
    
    def __init__(self, graph: Graph):
        self._graph = graph
    
    def __iter__(self):
        return iter(self._graph._graph.node_indices())
    
    def __len__(self):
        return self._graph.number_of_nodes()
    
    def __contains__(self, node):
        return self._graph._graph.has_node(node)
    
    def __getitem__(self, node):
        node_data = self._graph._graph.get_node_data(node)
        return node_data if isinstance(node_data, dict) else {}
    
    def __call__(self, data=False):
        """Return nodes with optional data."""
        if data:
            for idx in self._graph._graph.node_indices():
                node_data = self._graph._graph.get_node_data(idx)
                yield (idx, node_data if isinstance(node_data, dict) else {})
        else:
            for idx in self._graph._graph.node_indices():
                yield idx


class GraphEdgeView:
    """View of graph edges that mimics NetworkX EdgeView behavior."""
    
    def __init__(self, graph: Graph):
        self._graph = graph
    
    def __iter__(self):
        for edge_data in self._graph._graph.edge_list():
            src_idx, tgt_idx = edge_data
            yield (src_idx, tgt_idx)
    
    def __len__(self):
        return self._graph.number_of_edges()
    
    def __call__(self, data=False):
        """Return edges with optional data."""
        if data:
            for src_idx, tgt_idx in self._graph._graph.edge_list():
                edge_data = self._graph._graph.get_edge_data(src_idx, tgt_idx)
                yield (src_idx, tgt_idx, edge_data if isinstance(edge_data, dict) else {})
        else:
            for src_idx, tgt_idx in self._graph._graph.edge_list():
                yield (src_idx, tgt_idx)
    
    def __getitem__(self, edge):
        """Get edge data for a given edge (u, v)."""
        u, v = edge
        edge_data = self._graph._graph.get_edge_data(u, v)
        return edge_data if isinstance(edge_data, dict) else {}
    


from klotho.utils.algorithms.factors import to_factors
from typing import Union, List, Tuple, Dict, Set
from collections import namedtuple
from fractions import Fraction
import numpy as np
from sympy import Rational, root
import pandas as pd

A4_Hz   = 440.0
A4_MIDI = 69

from klotho.utils.data_structures.enums import DirectValueEnumMeta, Enum  

__all__ = [
    'ratio_to_cents',
    'cents_to_ratio',
    'cents_to_setclass',
    'ratio_to_setclass',
    'split_interval',
    'harmonic_mean',
    'arithmetic_mean',
    'logarithmic_distance',
    'interval_cost',
    'n_tet',
    'ratios_n_tet'
]

def ratio_to_cents(ratio: Union[int, float, Fraction, str], round_to: int = 4) -> float:
  '''
  Convert a musical interval ratio to cents, a logarithmic unit of measure.
  
  Args:
    ratio: The musical interval ratio as a string (e.g., '3/2') or float.
    
  Returns:
    The interval in cents as a float.
  '''
  # bad...
  # if isinstance(ratio, str):
  #   numerator, denominator = map(float, ratio.split('/'))
  # else:  # assuming ratio is already a float
  #   numerator, denominator = ratio, 1.0
  if isinstance(ratio, str):
    ratio = Fraction(ratio)
    numerator, denominator = ratio.numerator, ratio.denominator
  else:  # assuming ratio is already a float/int
    ratio = Fraction(ratio)
    numerator, denominator = ratio.numerator, ratio.denominator
  return round(1200 * np.log2(numerator / denominator), round_to)

def cents_to_ratio(cents: float) -> str:
  '''
  Convert a musical interval in cents to a ratio.
  
  Args:
    cents: The interval in cents to convert.
    
  Returns:
    The interval ratio as a float.
  '''
  return 2 ** (cents / 1200)

def cents_to_setclass(cent_value: float = 0.0, n_tet: int = 12, round_to: int = 2) -> float:
   return round((cent_value / 100)  % n_tet, round_to)

def ratio_to_setclass(ratio: Union[str, float], n_tet: int = 12, round_to: int = 2) -> float:
  '''
  Convert a musical interval ratio to a set class.
  
  Args:
    ratio: The musical interval ratio as a string (e.g., '3/2') or float.
    n_tet: The number of divisions in the octave, default is 12.
    round_to: The number of decimal places to round to, default is 2.
    
  Returns:
    The set class as a float.
  '''
  return cents_to_setclass(ratio_to_cents(ratio), n_tet, round_to)

def split_interval(interval:Union[int, float, Fraction, str], n:int = 2):
    '''
    Find the smallest sequence of n+1 integers that form n equal subdivisions of a given interval ratio.
    
    For a given interval ratio r and number of divisions n, finds a sequence of integers [a₀, a₁, ..., aₙ]
    where each adjacent pair forms the same ratio, and aₙ/a₀ equals the target ratio r.
    
    Algorithm:
    1. Initialize k = 1
    2. Loop:
        a. Calculate step size d = (r-1)k/n where r is target ratio
        b. If d is an integer:
            - Generate sequence [k, k+d, k+2d, ..., k+nd]
            - If sequence[n]/sequence[0] equals target ratio:
                return sequence and k
        c. Increment k
    
    Args:
        interval: The target interval ratio to be subdivided
        n: Number of equal subdivisions (default: 2)
        
    Returns:
        A named tuple containing:
            - sequence: List of n+1 integers forming the subdivisions
            - k: The smallest starting value that produces valid subdivisions
            
    Example:
        split_interval('3/2', 2) returns sequence [4, 5, 6] and k=4
        because 5/4 = 6/5 = √(3/2)
    '''
    result = namedtuple('result', ['sequence', 'k'])

    multiplier = Fraction(interval)
    k = 1
    while True:
        d = ((multiplier-1) * k) / n
        if d.denominator == 1:
            sequence = [k + i*int(d) for i in range(n+1)]
            if Fraction(sequence[-1], sequence[0]) == multiplier:
                return result(sequence, k)
        k += 1

def harmonic_mean(a: Union[int, float, Fraction, str], b: Union[int, float, Fraction, str]) -> Fraction:
    '''
    Calculate the harmonic mean between two values.
    
    The harmonic mean is defined as: 2 / (1/a + 1/b)
    
    Args:
        a: First value
        b: Second value
        
    Returns:
        The harmonic mean as a Fraction
    '''
    a, b = Fraction(a), Fraction(b)
    return 2 / (1/a + 1/b)

def arithmetic_mean(a: Union[int, float, Fraction, str], b: Union[int, float, Fraction, str]) -> Fraction:
    '''
    Calculate the arithmetic mean between two values.
    
    The arithmetic mean is defined as: (a + b) / 2
    
    Args:
        a: First value
        b: Second value
        
    Returns:
        The arithmetic mean as a Fraction
    '''
    a, b = Fraction(a), Fraction(b)
    return (a + b) / 2

def logarithmic_distance(a: Union[int, float, Fraction, str], b: Union[int, float, Fraction, str], 
                         equave: Union[int, float, Fraction, str] = 2) -> float:
    """
    Calculate the logarithmic distance between two musical intervals.
    
    Args:
        a: First interval
        b: Second interval
        equave: The equave to use for logarithmic scaling (default: 2 for octave)
        
    Returns:
        Logarithmic distance between the intervals
    """
    match a:
        case int() as i:
            r1 = Fraction(i, 1)
        case Fraction() as f:
            r1 = f
        case str() as s:
            r1 = Fraction(s)
        case _:
            raise TypeError("Unsupported type")

    match b:
        case int() as i:
            r2 = Fraction(i, 1)
        case Fraction() as f:
            r2 = f
        case str() as s:
            r2 = Fraction(s)
        case _:
            raise TypeError("Unsupported type")
            
    dist_interval = r2 / r1
    return abs(np.log(float(dist_interval)) / np.log(float(equave)))

def interval_cost(a: Union[int, float, Fraction, str], b: Union[int, float, Fraction, str], diff_coeff: float = 1.0, prime_coeff: float = 1.0,
                  equave: Union[int, float, Fraction, str] = 2) -> float:
    match a:
        case int() as i:
            r1 = Fraction(i, 1)
        case Fraction() as f:
            r1 = f
        case str() as s:
            r1 = Fraction(s)
        case _:
            raise TypeError("Unsupported type")

    match b:
        case int() as i:
            r2 = Fraction(i, 1)
        case Fraction() as f:
            r2 = f
        case str() as s:
            r2 = Fraction(s)
        case _:
            raise TypeError("Unsupported type")

    log_dist = logarithmic_distance(r1, r2, equave)

    f1 = to_factors(r1)
    f2 = to_factors(r2)
    p_all = set(f1.keys()) | set(f2.keys())
    prime_diff = sum(abs(f1.get(p, 0) - f2.get(p, 0)) for p in p_all)

    return diff_coeff * log_dist + prime_coeff * prime_diff

def n_tet(divisions: int = 12, equave: Union[int, float, Fraction, str] = 2, nth_division: int = 1, symbolic: bool = False) -> Union[float, Rational]:
    '''
    Calculate the size of the nth division of an interval in equal temperament.
    
    Args:
        divisions: The number of equal divisions (default: 12)
        equave: The interval to divide (default: 2 for octave)
        nth_division: The nth division to calculate (default: 1)
        symbolic: If True, return symbolic expression instead of float (default: False)
    
    Returns:
        The frequency ratio either as a float or as a sympy expression
    '''
    ratio = root(Fraction(equave), Rational(divisions)) ** nth_division
    return ratio if symbolic else float(ratio)

def ratios_n_tet(divisions: int = 12, equave: Union[int, float, Fraction, str] = 2, symbolic: bool = False) -> List[Union[float, Rational]]:
  '''
  Calculate the ratios of the divisions of an interval in equal temperament.

  see:  https://en.wikipedia.org/wiki/Equal_temperament

  Args:
    divisions: The number of equal divisions
    equave: The interval to divide (default is 2 for an octave)
    symbolic: If True, return symbolic expression instead of float (default: False)
    
  Returns:
    A list of the frequency ratios of the divisions
  '''
  return [n_tet(divisions, equave, nth_division, symbolic) for nth_division in range(divisions)]



import mido
from typing import Union, List
from pathlib import Path

def midi(obj: Union['TemporalUnit', 'TemporalUnitSequence', 'TemporalBlock'], 
         file_path: Union[str, Path], 
         ticks_per_beat: int = 480,
         default_velocity: int = 64,
         middle_c: int = 60) -> None:
    
    file_path = Path(file_path)
    if file_path.suffix != '.mid':
        file_path = file_path.with_suffix('.mid')
    
    mid = mido.MidiFile(ticks_per_beat=ticks_per_beat)
    
    if hasattr(obj, 'events'):
        _add_temporal_unit_track(mid, obj, default_velocity, middle_c, ticks_per_beat)
    elif hasattr(obj, 'seq'):
        _add_temporal_sequence_track(mid, obj, default_velocity, middle_c, ticks_per_beat)
    elif hasattr(obj, 'rows'):
        _add_temporal_block_tracks(mid, obj, default_velocity, middle_c, ticks_per_beat)
    else:
        raise ValueError(f"Unsupported object type: {type(obj)}")
    
    mid.save(file_path)

def _add_temporal_unit_track(mid: mido.MidiFile, 
                           temporal_unit: 'TemporalUnit',
                           default_velocity: int,
                           middle_c: int,
                           ticks_per_beat: int) -> None:
    
    track = mido.MidiTrack()
    mid.tracks.append(track)
    
    events = temporal_unit.events
    absolute_offset = temporal_unit.offset
    
    previous_time = 0
    
    for _, event in events.iterrows():
        if event['is_rest']:
            continue
            
        start_time = event['start'] + absolute_offset
        duration = abs(event['duration'])
        
        note_on_delta = _seconds_to_ticks(start_time - previous_time, ticks_per_beat, 120)
        note_off_delta = _seconds_to_ticks(duration, ticks_per_beat, 120)
        
        track.append(mido.Message('note_on', 
                                 channel=0, 
                                 note=middle_c, 
                                 velocity=default_velocity, 
                                 time=note_on_delta))
        
        track.append(mido.Message('note_off', 
                                 channel=0, 
                                 note=middle_c, 
                                 velocity=0, 
                                 time=note_off_delta))
        
        previous_time = start_time + duration

def _add_temporal_sequence_track(mid: mido.MidiFile,
                               temporal_sequence: 'TemporalUnitSequence',
                               default_velocity: int,
                               middle_c: int,
                               ticks_per_beat: int) -> None:
    
    track = mido.MidiTrack()
    mid.tracks.append(track)
    
    absolute_offset = temporal_sequence.offset
    previous_time = 0
    
    for temporal_unit in temporal_sequence.seq:
        events = temporal_unit.events
        unit_offset = temporal_unit.offset
        
        for _, event in events.iterrows():
            if event['is_rest']:
                continue
                
            start_time = event['start'] + unit_offset + absolute_offset
            duration = abs(event['duration'])
            
            note_on_delta = _seconds_to_ticks(start_time - previous_time, ticks_per_beat, 120)
            note_off_delta = _seconds_to_ticks(duration, ticks_per_beat, 120)
            
            track.append(mido.Message('note_on', 
                                     channel=0, 
                                     note=middle_c, 
                                     velocity=default_velocity, 
                                     time=note_on_delta))
            
            track.append(mido.Message('note_off', 
                                     channel=0, 
                                     note=middle_c, 
                                     velocity=0, 
                                     time=note_off_delta))
            
            previous_time = start_time + duration

def _add_temporal_block_tracks(mid: mido.MidiFile,
                             temporal_block: 'TemporalBlock',
                             default_velocity: int,
                             middle_c: int,
                             ticks_per_beat: int) -> None:
    
    absolute_offset = temporal_block.offset
    
    for row_idx, row in enumerate(temporal_block.rows):
        track = mido.MidiTrack()
        mid.tracks.append(track)
        
        previous_time = 0
        
        if hasattr(row, 'events'):
            events = row.events
            unit_offset = row.offset
            
            for _, event in events.iterrows():
                if event['is_rest']:
                    continue
                    
                start_time = event['start'] + unit_offset + absolute_offset
                duration = abs(event['duration'])
                
                note_on_delta = _seconds_to_ticks(start_time - previous_time, ticks_per_beat, 120)
                note_off_delta = _seconds_to_ticks(duration, ticks_per_beat, 120)
                
                track.append(mido.Message('note_on', 
                                         channel=0, 
                                         note=middle_c + row_idx, 
                                         velocity=default_velocity, 
                                         time=note_on_delta))
                
                track.append(mido.Message('note_off', 
                                         channel=0, 
                                         note=middle_c + row_idx, 
                                         velocity=0, 
                                         time=note_off_delta))
                
                previous_time = start_time + duration
                
        elif hasattr(row, 'seq'):
            for temporal_unit in row.seq:
                events = temporal_unit.events
                unit_offset = temporal_unit.offset
                
                for _, event in events.iterrows():
                    if event['is_rest']:
                        continue
                        
                    start_time = event['start'] + unit_offset + absolute_offset
                    duration = abs(event['duration'])
                    
                    note_on_delta = _seconds_to_ticks(start_time - previous_time, ticks_per_beat, 120)
                    note_off_delta = _seconds_to_ticks(duration, ticks_per_beat, 120)
                    
                    track.append(mido.Message('note_on', 
                                             channel=0, 
                                             note=middle_c + row_idx, 
                                             velocity=default_velocity, 
                                             time=note_on_delta))
                    
                    track.append(mido.Message('note_off', 
                                             channel=0, 
                                             note=middle_c + row_idx, 
                                             velocity=0, 
                                             time=note_off_delta))
                    
                    previous_time = start_time + duration
                    
        elif hasattr(row, 'rows'):
            _add_temporal_block_tracks_recursive(track, row, absolute_offset, 
                                               default_velocity, middle_c + row_idx, 
                                               ticks_per_beat, previous_time)

def _add_temporal_block_tracks_recursive(track: mido.MidiTrack,
                                       temporal_block: 'TemporalBlock',
                                       absolute_offset: float,
                                       default_velocity: int,
                                       note_offset: int,
                                       ticks_per_beat: int,
                                       previous_time: float) -> float:
    
    block_offset = temporal_block.offset
    
    for sub_row_idx, sub_row in enumerate(temporal_block.rows):
        if hasattr(sub_row, 'events'):
            events = sub_row.events
            unit_offset = sub_row.offset
            
            for _, event in events.iterrows():
                if event['is_rest']:
                    continue
                    
                start_time = event['start'] + unit_offset + block_offset + absolute_offset
                duration = abs(event['duration'])
                
                note_on_delta = _seconds_to_ticks(start_time - previous_time, ticks_per_beat, 120)
                note_off_delta = _seconds_to_ticks(duration, ticks_per_beat, 120)
                
                track.append(mido.Message('note_on', 
                                         channel=0, 
                                         note=note_offset + sub_row_idx, 
                                         velocity=default_velocity, 
                                         time=note_on_delta))
                
                track.append(mido.Message('note_off', 
                                         channel=0, 
                                         note=note_offset + sub_row_idx, 
                                         velocity=0, 
                                         time=note_off_delta))
                
                previous_time = start_time + duration
                
        elif hasattr(sub_row, 'seq'):
            for temporal_unit in sub_row.seq:
                events = temporal_unit.events
                unit_offset = temporal_unit.offset
                
                for _, event in events.iterrows():
                    if event['is_rest']:
                        continue
                        
                    start_time = event['start'] + unit_offset + block_offset + absolute_offset
                    duration = abs(event['duration'])
                    
                    note_on_delta = _seconds_to_ticks(start_time - previous_time, ticks_per_beat, 120)
                    note_off_delta = _seconds_to_ticks(duration, ticks_per_beat, 120)
                    
                    track.append(mido.Message('note_on', 
                                             channel=0, 
                                             note=note_offset + sub_row_idx, 
                                             velocity=default_velocity, 
                                             time=note_on_delta))
                    
                    track.append(mido.Message('note_off', 
                                             channel=0, 
                                             note=note_offset + sub_row_idx, 
                                             velocity=0, 
                                             time=note_off_delta))
                    
                    previous_time = start_time + duration
    
    return previous_time

def _seconds_to_ticks(seconds: float, ticks_per_beat: int, bpm: int) -> int:
    if seconds <= 0:
        return 0
    beats_per_second = bpm / 60.0
    beats = seconds * beats_per_second
    ticks = int(beats * ticks_per_beat)
    return max(0, ticks) 


from abc import ABC, abstractmethod
from typing import Any, Dict, Union, Tuple
import numpy as np

class FieldFunction(ABC):
    '''Base class for field functions with variable input dimensionality and output range.'''

    def __init__(self, input_dim: int = 1, output_dim: int = 1, 
                 output_range: Tuple[float, float] = (float('-inf'), float('inf')),
                 parameters: Dict[str, Any] = None):
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.output_range = output_range
        self.parameters = parameters or {}

    @abstractmethod
    def _raw_function(self, x: np.ndarray) -> np.ndarray:
        '''Raw function to be implemented by subclasses.'''
        pass

    def __call__(self, x: Union[float, list, np.ndarray]) -> Union[float, np.ndarray]:
        '''Apply the function to input x, ensuring correct dimensionality and range.'''
        x = np.atleast_1d(x)
        if x.shape[-1] != self.input_dim:
            raise ValueError(f"Input dimensionality {x.shape[-1]} does not match expected {self.input_dim}")
        
        result = self._raw_function(x)
        
        if self.output_dim > 1:
            result = result.reshape(-1, self.output_dim)
        
        return np.clip(result, self.output_range[0], self.output_range[1])

    @staticmethod
    def compose(*functions: 'FieldFunction') -> 'FieldFunction':
        '''Compose multiple FieldFunctions.'''
        if len(functions) < 2:
            raise ValueError("At least two functions are required for composition.")
        
        for i in range(len(functions) - 1):
            if functions[i].input_dim != functions[i+1].output_dim:
                raise ValueError(f"Output dimension of function {i+1} must match input dimension of function {i}.")
        
        class ComposedFunction(FieldFunction):
            def __init__(self, *funcs):
                super().__init__(funcs[-1].input_dim, funcs[0].output_dim, funcs[0].output_range)
                self.functions = funcs
            
            def _raw_function(self, x):
                for func in reversed(self.functions):
                    x = func._raw_function(x)
                return x
        
        return ComposedFunction(*functions)

    def set_parameters(self, **kwargs):
        '''Set parameters of the function.'''
        self.parameters.update(kwargs)

# -------------------------------------------

class Identity(FieldFunction):
    '''Identity function.'''
    def _raw_function(self, x: np.ndarray) -> np.ndarray:
        return x

class Sigmoid(FieldFunction):
    '''Sigmoid function.'''
    def __init__(self, input_dim: int = 1):
        super().__init__(input_dim, 1, (0, 1))
    
    def _raw_function(self, x: np.ndarray) -> np.ndarray:
        return 1 / (1 + np.exp(-x))

class Gaussian(FieldFunction):
    '''Gaussian function.'''
    def __init__(self, input_dim: int = 1, mu: float = 0, sigma: float = 1):
        super().__init__(input_dim, 1, (0, 1), {'mu': mu, 'sigma': sigma})
    
    def _raw_function(self, x: np.ndarray) -> np.ndarray:
        return np.exp(-((x - self.parameters['mu']) ** 2) / (2 * self.parameters['sigma'] ** 2))

class Polynomial(FieldFunction):
    '''Polynomial function.'''
    def __init__(self, coefficients: list, input_dim: int = 1):
        super().__init__(input_dim, 1, parameters={'coefficients': coefficients})
    
    def _raw_function(self, x: np.ndarray) -> np.ndarray:
        return sum(c * x**i for i, c in enumerate(self.parameters['coefficients']))



"""
Chronos: A specialized module for working with time and rhythm in music.

The word "chronos" originates from Ancient Greek and is deeply rooted in both language 
and mythology. In Greek, "χρόνος" (chronos) means "time".

In Greek mythology, Chronos (not to be confused with Cronus, the Titan) is personified as 
the god of time. His representation often symbolizes the endless passage of time and the 
cycles of creation and destruction within the universe.
"""
from . import rhythm_pairs
from . import rhythm_trees
from . import temporal_units
from . import utils

from .rhythm_pairs import RhythmPair
from .rhythm_trees import RhythmTree, Meas
from .temporal_units import TemporalUnit, TemporalUnitSequence, TemporalBlock

from .utils.beat import *
from .utils.tempo import *
from .utils.time_conversion import *
from .utils.beat import __all__ as beat_all
from .utils.tempo import __all__ as tempo_all
from .utils.time_conversion import __all__ as time_conversion_all

__all__ = [
    # Modules
    'rhythm_pairs', 
    'rhythm_trees', 
    'temporal_units', 
    'utils',
    
    # Classes
    'RhythmPair',
    'RhythmTree',
    'Meas',
    'TemporalUnit',
    'TemporalUnitSequence',
    'TemporalBlock',
]

__all__.extend(beat_all)
__all__.extend(tempo_all)
__all__.extend(time_conversion_all)



# ------------------------------------------------------------------------------------
# Klotho/klotho/chronos/rhythm_trees/rt.py
# ------------------------------------------------------------------------------------
'''
--------------------------------------------------------------------------------------
A rhythm tree (RT) is a list representing a rhythmic structure. This list is organized 
hierarchically in sub lists , just as time is organized in measures, time signatures, 
pulses and rhythmic elements in the traditional notation.

Hence, the expression form of rhythm trees is crucially different from that of onsets 
and offsets. It can be exacting and not very "ergonomic", from a musician's point of 
view : rhythm trees can be long, with a great number of parenthesis and sub lists 
nested within each others.

see: https://support.ircam.fr/docs/om/om6-manual/co/RT.html
--------------------------------------------------------------------------------------
'''
from fractions import Fraction
from typing import Union, Tuple
from tabulate import tabulate

from klotho.topos.graphs import Tree
from klotho.utils.algorithms.groups import print_subdivisions
from klotho.utils.data_structures import Group
from .meas import Meas
from .algorithms import sum_proportions, measure_complexity, ratios_to_subdivs
from ..utils.beat import calc_onsets


class RhythmTree(Tree):
    '''
    A rhythm tree is a list representing a rhythmic structure. This list is organized 
    hierarchically in sub lists, just as time is organized in measures, time signatures, 
    pulses and rhythmic elements in the traditional notation.

    Traditionally, rhythm is broken up into several data : meter, measure(s) and duration(s). 
    Rhythm trees must enclose these information in lists and sub list.

    This elementary rhythm:

    [1/4, 1/4, 1/4, 1/4] --> (four 1/4-notes in 4/4 time)

    can be expressed as follows :

    ( ? ( (4//4 (1 1 1 1) ) ) )

    A tree structure can be reduced to a list : (D (S)).


    >> Main Components : Duration and Subdivisions

    D = a duration , or number of measures : ( ? ) or a number ( n ).
    When D = ?, OM calculates the duration.
    By default, this duration is equal to 1.

    S = subdivisions (S) of this duration, that is a time signature and rhythmic proportions.
    Time signature = n // n   or ( n n ).
    It must be specified at each new measure, even if it remains unchanged.

    Rhythm = proportions : ( n n n n )

    see: https://support.ircam.fr/docs/om/om6-manual/co/RT1.html
    '''
    def __init__(self, 
                 span:int                      = 1,
                 meas:Union[Meas,Fraction,str] = '1/1',
                 subdivisions:Tuple            = (1,1)):
        
        super().__init__(Meas(meas).numerator * span, subdivisions)
        
        self._meta['span'] = span
        self._meta['meas'] = str(Meas(meas))
        self._meta['type'] = None
        self._list = Group((Meas(meas).numerator * span, self._cast_subdivs(subdivisions)))
        
        self._evaluate()

    @classmethod
    def from_tree(cls, tree:Tree, span:int = 1):
        return cls(span = span, meas = Meas(tree[tree.root]['metric_duration']), subdivisions = tree.group.S)
    
    @classmethod
    def from_ratios(cls, ratios:Tuple[Fraction, float, str], span:int = 1):
        ratios = tuple(Fraction(r) for r in ratios)
        S = ratios_to_subdivs(ratios)
        meas = Meas(sum(abs(r) for r in ratios))
        return cls(span = span, meas = meas, subdivisions = S)

    @property
    def span(self):
        return self._meta['span']

    @property
    def meas(self):
        return Meas(self._meta['meas'])

    @property
    def subdivisions(self):
        return self._list.S

    def _cast_subdivs(self, children):
        def convert_to_tuple(item):
            if isinstance(item, RhythmTree):
                return (item.meas.numerator * item.span, item.subdivisions)
            if isinstance(item, tuple):
                return tuple(convert_to_tuple(x) for x in item)
            return item
        
        return tuple(convert_to_tuple(child) for child in children)

    def _build_subdivisions(self, root_node=None):
        """Build subdivisions structure from the current graph state.
        
        Parameters
        ----------
        root_node : int, optional
            The node to start building from (default: self.root)
            
        Returns
        -------
        tuple
            Nested tuple structure representing subdivisions
        """
        if root_node is None:
            root_node = self.root
        
        def get_node_value(node):
            return self[node].get('proportion', self[node].get('label', 1))
        
        def get_children(node):
            return list(self.successors(node))
        
        return self._build_nested_structure(root_node, get_node_value, get_children)
    
    def _update_group_structure(self):
        """Update the Group structure, preserving D and updating S."""
        if hasattr(self, '_list'):
            new_subdivisions = self._build_subdivisions()
            if isinstance(new_subdivisions, tuple) and len(new_subdivisions) > 1:
                new_s = new_subdivisions[1]
            else:
                new_s = (1,)
            self._list = Group((self._list.D, new_s))
    
    @property
    def durations(self):
        return tuple(self.nodes[n]['metric_duration'] for n in self.leaf_nodes)
    
    @property
    def onsets(self):
        return tuple(self.nodes[n]['metric_onset'] for n in self.leaf_nodes)
    
    @property
    def info(self):
        ordered_meta = {k: self._meta[k] for k in ['span', 'meas', 'type']}
        ordered_meta['depth'] = self.depth
        ordered_meta['k'] = self.k
        meta_str = ' | '.join(f"{k}: {v}" for k, v in ordered_meta.items())
        
        table_data = [
            [str(r) for r in self.durations],
            [str(o) for o in self.onsets]
        ]
        
        duration_onset_table = tabulate(
            table_data,
            headers=[],
            tablefmt='plain'
        )
        
        table_lines = duration_onset_table.split('\n')
        durations_line = f"Durations: {table_lines[0]}"
        onsets_line = f"Onsets:    {table_lines[1]}"
        
        content = [
            meta_str,
            f"Subdivs: {print_subdivisions(self.subdivisions)}",
            onsets_line,
            durations_line
        ]
        
        width = max(len(line) for line in content)
        border = '-' * width
        
        return (
            f"{border}\n"
            f"{content[0]}\n"
            f"{border}\n"
            f"{content[1]}\n"
            f"{border}\n"
            f"{content[2]}\n"
            f"{content[3]}\n"
            f"{border}\n"
        )
    
    # @property
    # def type(self):
    #     if self._meta['type'] is None:
    #         self._meta['type'] = self._set_type()
    #     return self._meta['type']
    
    def _evaluate(self):
        """
        Evaluate the rhythm tree to compute metric durations and onsets.
        
        This method processes the tree in two phases:
        1. Computation of metric durations and proportions for all nodes
        2. Computation of metric onsets based on durations
        """
        self[self.root]['metric_duration'] = self.meas * self.span
        
        def _process_child_durations(child, div, parent_ratio, parent_is_negative=False):
            """
            Process duration and proportion for a single child node.
            
            Parameters
            ----------
            child : int
                Child node ID to process.
            div : int
                Sum of all proportions at this level.
            parent_ratio : Fraction
                Parent node's metric duration ratio.
            parent_is_negative : bool, optional
                Whether the parent node has a negative proportion.
            """
            child_data = self[child]
            
            s = child_data['label']
            if 'meta' in child_data:
                s = s * child_data['meta']['span']
            s = int(s) if isinstance(s, float) else s
            
            if parent_is_negative and s > 0:
                s = -s
            
            ratio = Fraction(s, div) * parent_ratio
            if s < 0:
                ratio = -abs(ratio)
            self[child]['metric_duration'] = ratio
            self[child]['proportion'] = s
            if self.out_degree(child) > 0:
                _process_subtree(child, ratio)
            self[child].pop('label', None)
        
        def _process_subtree(node=0, parent_ratio=self.span * self.meas.to_fraction()):
            """
            Process a subtree to compute metric durations and proportions.
            
            Parameters
            ----------
            node : int, optional
                Root node of subtree to process (default is 0).
            parent_ratio : Fraction, optional
                Parent node's metric duration ratio.
            """
            node_data = self[node]
            
            if 'meta' in node_data:
                node_data['label'] = node_data['label'] * node_data['meta']['span']
            
            label = node_data['label']
            is_tied = isinstance(label, float)
            self[node]['tied'] = is_tied
            label_value = int(label) if is_tied else label
            
            self[node]['proportion'] = label_value
            children = list(self.successors(node))
            
            if not children:
                ratio = Fraction(label_value) * parent_ratio
                self[node]['metric_duration'] = ratio
                self[node].pop('label', None)
                return
            
            div = int(sum(abs(self[c]['label'] * 
                             self[c]['meta']['span'] if 'meta' in self[c]
                             else self[c]['label']) 
                         for c in children))
            
            node_is_negative = label_value < 0
                        
            for child in children:
                _process_child_durations(child, div, parent_ratio, node_is_negative)
            
            self[node].pop('label', None)
        
        _process_subtree()
        
        leaf_durations = [self[n]['metric_duration'] for n in self.leaf_nodes]
        leaf_onsets = calc_onsets(leaf_durations)
        
        for n, o in zip(self.leaf_nodes, leaf_onsets):
            self[n]['metric_onset'] = o
        
        for node in reversed(list(self.topological_sort())):
            if self.out_degree(node) > 0:
                children = list(self.successors(node))
                leftmost_child = children[0]
                self[node]['metric_onset'] = self[leftmost_child]['metric_onset']

    def _set_type(self):
        div = sum_proportions(self.subdivisions)
        if bin(div).count('1') != 1 and div != self.meas.numerator:
            return 'complex'
        return 'complex' if measure_complexity(self.subdivisions) else 'simple'

    def __len__(self):
        return len(self.durations)

    def __str__(self):
        return f"RhythmTree(span={self.span}, meas={self.meas}, subdivisions={print_subdivisions(self.subdivisions)})"

    def __repr__(self):
        return self.__str__()
    
    def subtree(self, node, renumber=True):
        """Extract a rhythm subtree rooted at the given node.
        
        The subtree becomes a new RhythmTree with:
        - span = 1
        - meas = metric_duration of the selected node
        - subdivisions = reconstructed from the subtree structure
        
        Parameters
        ----------
        node : int
            The node to use as the root of the subtree
        renumber : bool, optional
            Whether to renumber the nodes in the new tree (default: True)
            
        Returns
        -------
        RhythmTree
            A new RhythmTree representing the subtree
        """
        if node not in self:
            raise ValueError(f"Node {node} not found in tree")

        subdivisions_structure = self._build_subdivisions(node)
        if isinstance(subdivisions_structure, tuple) and len(subdivisions_structure) > 1:
            subdivisions = subdivisions_structure[1]
        else:
            subdivisions = (1,)
        
        node_duration = self[node].get('metric_duration')
        if node_duration is None:
            meas = '1/1'
        else:
            meas = Meas(node_duration)
        
        new_rt = RhythmTree(span=1, meas=meas, subdivisions=subdivisions)
        
        if renumber:
            new_rt.renumber_nodes()
        
        return new_rt
    
    def graft_subtree(self, target_node, subtree, mode='replace'):
        """Graft a subtree and re-evaluate the rhythm tree."""
        result = super().graft_subtree(target_node, subtree, mode)
        
        for node in self.nodes:
            node_data = self[node]
            if 'label' not in node_data:
                node_data['label'] = node_data.get('proportion', 1)
        
        self._evaluate()
        return result

    def make_rest(self, node):
        """
        Make a node and all its descendants into rests by setting their proportions to negative.
        
        Parameters
        ----------
        node : int
            The node ID to make into a rest along with all its descendants.
            
        Raises
        ------
        ValueError
            If the node is not found in the tree.
        """
        if node not in self:
            raise ValueError(f"Node {node} not found in tree")
        
        descendants_to_modify = [node] + list(self.descendants(node))
        
        for n in descendants_to_modify:
            node_data = self[n]
            if 'proportion' in node_data and node_data['proportion'] > 0:
                node_data['proportion'] = -abs(node_data['proportion'])
                node_data['metric_duration'] = -abs(node_data['metric_duration'])
        
        self._update_group_structure()



from typing import List, Callable, Optional
import random
from ...topos.graphs import Graph

__all__ = [
    'minimum_cost_path',
    'greedy_random_walk',
    'probabilistic_random_walk',
    'deterministic_greedy_walk',
    'prim_order_traversal',
    'greedy_nearest_unvisited',
    'dijkstra_order_traversal',
    'weighted_dfs_traversal'
]

def greedy_tsp(G: Graph, source=None, **kwargs) -> List[int]:
    """
    Simple greedy TSP approximation algorithm.
    
    Parameters
    ----------
    G : Graph
        Weighted graph
    source : node, optional
        Starting node. If None, uses first node.
    **kwargs
        Additional parameters (ignored)
        
    Returns
    -------
    List[int]
        Approximate TSP tour
    """
    if not G.nodes:
        return []
    
    if source is None:
        source = next(iter(G.nodes))
    
    if source not in G:
        raise ValueError(f"Source node {source} not in graph")
    
    unvisited = set(G.nodes) - {source}
    tour = [source]
    current = source
    
    while unvisited:
        min_weight = float('inf')
        next_node = None
        
        for node in unvisited:
            if G.has_edge(current, node):
                try:
                    weight = G[current][node].get('weight', 1.0)
                    if weight < min_weight:
                        min_weight = weight
                        next_node = node
                except (KeyError, TypeError):
                    if 1.0 < min_weight:
                        min_weight = 1.0
                        next_node = node
        
        if next_node is None:
            # No direct edge, find nearest unvisited node
            next_node = unvisited.pop()
        else:
            unvisited.remove(next_node)
        
        tour.append(next_node)
        current = next_node
    
    return tour

def minimum_cost_path(
    G: Graph,
    traversal_func: Optional[Callable] = None,
    **kwargs
) -> List[int]:
    """
    Find a minimum cost path through a weighted graph using flexible traversal functions.
    
    Delegates to the specified traversal function, providing sensible defaults
    for missing required parameters. The default traversal function is greedy_tsp
    which solves the traveling salesman problem.

    Parameters
    ----------
    G : Graph
        Weighted graph with numeric edge weights representing costs.
    traversal_func : Callable, optional
        Function to use for graph traversal. Receives the graph as first
        argument and all other parameters via kwargs. Defaults to greedy_tsp.
    **kwargs
        All parameters for the traversal function. If using the default
        greedy_tsp and 'source' is not provided, defaults to the first node.

    Returns
    -------
    List[int]
        Ordered list of node indices representing the path found by the
        traversal function.

    Raises
    ------
    ValueError
        If required nodes are not in the graph or if no valid path exists.

    Examples
    --------
    Use default greedy TSP with automatic source:
    
    >>> from klotho.topos.graphs import Graph
    >>> G = Graph.digraph()
    >>> G.add_edge(0, 1, weight=1)
    >>> G.add_edge(1, 2, weight=2) 
    >>> G.add_edge(0, 2, weight=4)
    >>> path = minimum_cost_path(G)  # Uses first node as source
    
    Use default greedy TSP with specified source:
    
    >>> path = minimum_cost_path(G, source=0)
    
    Use custom traversal function:
    
    >>> def custom_dfs(graph, source, depth_limit=None):
    ...     return list(graph.descendants(source))[:depth_limit]
    >>> path = minimum_cost_path(G, traversal_func=custom_dfs, source=0, depth_limit=2)

    Notes
    -----
    This function provides maximum flexibility by accepting any traversal function
    and passing all parameters via kwargs. When no traversal function is specified,
    it uses a greedy TSP algorithm which finds an approximate solution
    to the traveling salesman problem.
    """
    if traversal_func is None:
        return greedy_tsp(G, **kwargs)
    
    return traversal_func(G, **kwargs)

def greedy_random_walk(G, source, steps: int = 10, weight: str = 'weight', 
                      target: Optional[int] = None, **kwargs) -> List[int]:
    """
    Perform a greedy walk choosing minimum weight edges with random tie-breaking.
    
    Parameters
    ----------
    G : networkx.Graph
        The graph to walk on
    source : node
        Starting node
    steps : int, optional
        Maximum number of steps to take (default: 10)
    weight : str, optional
        Edge attribute to use for decision making (default: 'weight')
    target : node, optional
        If provided, stop early when target is reached
    **kwargs
        Additional parameters (ignored)
        
    Returns
    -------
    List[int]
        Path as list of nodes visited
    """
    if source not in G:
        raise ValueError(f"Source node {source} not in graph")
    
    path = [source]
    current = source
    
    for step in range(steps):
        neighbors = list(G.neighbors(current))
        
        if not neighbors:
            break
            
        neighbor_weights = []
        for neighbor in neighbors:
            try:
                edge_weight = G[current][neighbor].get(weight, 1.0)
                neighbor_weights.append((neighbor, edge_weight))
            except (KeyError, TypeError):
                neighbor_weights.append((neighbor, 1.0))
        
        min_weight = min(neighbor_weights, key=lambda x: x[1])[1]
        
        min_weight_neighbors = [neighbor for neighbor, w in neighbor_weights if w == min_weight]
        
        next_node = random.choice(min_weight_neighbors)
        path.append(next_node)
        current = next_node
        
        if target is not None and current == target:
            break
    
    return path


def probabilistic_random_walk(G, source, steps: int = 10, weight: str = 'weight',
                             target: Optional[int] = None, inverse_weights: bool = True, 
                             **kwargs) -> List[int]:
    """
    Perform a probabilistic walk where lower weights have higher probability.
    
    Parameters
    ----------
    G : networkx.Graph
        The graph to walk on
    source : node
        Starting node
    steps : int, optional
        Maximum number of steps to take (default: 10)
    weight : str, optional
        Edge attribute to use for decision making (default: 'weight')
    target : node, optional
        If provided, stop early when target is reached
    inverse_weights : bool, optional
        If True, lower weights get higher probability (default: True)
    **kwargs
        Additional parameters (ignored)
        
    Returns
    -------
    List[int]
        Path as list of nodes visited
    """
    if source not in G:
        raise ValueError(f"Source node {source} not in graph")
    
    path = [source]
    current = source
    
    for step in range(steps):
        neighbors = list(G.neighbors(current))
        
        if not neighbors:
            break
            
        weights = []
        for neighbor in neighbors:
            try:
                edge_weight = G[current][neighbor].get(weight, 1.0)
                weights.append(edge_weight)
            except (KeyError, TypeError):
                weights.append(1.0)
        
        if inverse_weights:
            inv_weights = [1.0 / max(w, 1e-10) for w in weights]
            total = sum(inv_weights)
            probabilities = [w / total for w in inv_weights]
        else:
            total = sum(weights)
            probabilities = [w / total for w in weights] if total > 0 else [1.0/len(weights)] * len(weights)
        
        next_node = random.choices(neighbors, weights=probabilities)[0]
        path.append(next_node)
        current = next_node
        
        if target is not None and current == target:
            break
    
    return path


def deterministic_greedy_walk(G, source, steps: int = 10, weight: str = 'weight',
                             target: Optional[int] = None, **kwargs) -> List[int]:
    """
    Always choose the neighbor with minimum weight (deterministic, no randomness).
    
    Parameters
    ----------
    G : networkx.Graph
        The graph to walk on
    source : node
        Starting node
    steps : int, optional
        Maximum number of steps to take (default: 10)
    weight : str, optional
        Edge attribute to use for decision making (default: 'weight')
    target : node, optional
        If provided, stop early when target is reached
    **kwargs
        Additional parameters (ignored)
        
    Returns
    -------
    List[int]
        Path as list of nodes visited
    """
    if source not in G:
        raise ValueError(f"Source node {source} not in graph")
    
    path = [source]
    current = source
    visited = set([source])
    
    for step in range(steps):
        neighbors = [n for n in G.neighbors(current) if n not in visited]
        
        if not neighbors:
            neighbors = list(G.neighbors(current))
            if not neighbors:
                break
        
        min_neighbor = None
        min_weight = float('inf')
        
        for neighbor in neighbors:
            try:
                edge_weight = G[current][neighbor].get(weight, 1.0)
                if edge_weight < min_weight:
                    min_weight = edge_weight
                    min_neighbor = neighbor
            except (KeyError, TypeError):
                if 1.0 < min_weight:
                    min_weight = 1.0
                    min_neighbor = neighbor
        
        if min_neighbor is None:
            break
            
        path.append(min_neighbor)
        current = min_neighbor
        visited.add(current)
        
        if target is not None and current == target:
            break
    
    return path

def prim_order_traversal(G, source, weight: str = 'weight', **kwargs) -> List[int]:
    """
    Visit all nodes in the order they would be added by Prim's MST algorithm.
    
    This guarantees visiting every node while prioritizing edges with lower weights.
    Follows the minimum spanning tree construction order.
    
    Parameters
    ----------
    G : networkx.Graph
        The graph to traverse
    source : node
        Starting node for traversal
    weight : str, optional
        Edge attribute to use for weights (default: 'weight')
    **kwargs
        Additional parameters (ignored)
        
    Returns
    -------
    List[int]
        Nodes visited in Prim's algorithm order
    """
    if source not in G:
        raise ValueError(f"Source node {source} not in graph")
    
    visited = set([source])
    path = [source]
    
    while len(visited) < G.number_of_nodes():
        min_edge = None
        min_weight = float('inf')
        next_node = None
        
        for node in visited:
            for neighbor in G.neighbors(node):
                if neighbor not in visited:
                    try:
                        edge_weight = G[node][neighbor].get(weight, 1.0)
                        if edge_weight < min_weight:
                            min_weight = edge_weight
                            min_edge = (node, neighbor)
                            next_node = neighbor
                    except (KeyError, TypeError):
                        if 1.0 < min_weight:
                            min_weight = 1.0
                            min_edge = (node, neighbor)
                            next_node = neighbor
        
        if next_node is None:
            break
            
        visited.add(next_node)
        path.append(next_node)
    
    return path


def greedy_nearest_unvisited(G, source, weight: str = 'weight', **kwargs) -> List[int]:
    """
    Visit all nodes by always moving to the nearest unvisited neighbor.
    
    At each step, chooses the unvisited neighbor with the minimum edge weight.
    If no unvisited neighbors exist, backtracks or jumps to nearest unvisited node.
    
    Parameters
    ----------
    G : networkx.Graph
        The graph to traverse
    source : node
        Starting node for traversal
    weight : str, optional
        Edge attribute to use for weights (default: 'weight')
    **kwargs
        Additional parameters (ignored)
        
    Returns
    -------
    List[int]
        Nodes visited in greedy nearest order
    """
    if source not in G:
        raise ValueError(f"Source node {source} not in graph")
    
    visited = set([source])
    path = [source]
    current = source
    
    while len(visited) < G.number_of_nodes():
        min_neighbor = None
        min_weight = float('inf')
        
        for neighbor in G.neighbors(current):
            if neighbor not in visited:
                try:
                    edge_weight = G[current][neighbor].get(weight, 1.0)
                    if edge_weight < min_weight:
                        min_weight = edge_weight
                        min_neighbor = neighbor
                except (KeyError, TypeError):
                    if 1.0 < min_weight:
                        min_weight = 1.0
                        min_neighbor = neighbor
        
        if min_neighbor is not None:
            visited.add(min_neighbor)
            path.append(min_neighbor)
            current = min_neighbor
        else:
            unvisited = set(G.nodes()) - visited
            if not unvisited:
                break
                
            min_distance = float('inf')
            best_next = None
            best_path_node = None
            
            for visited_node in visited:
                for unvisited_node in unvisited:
                    if G.has_edge(visited_node, unvisited_node):
                        try:
                            edge_weight = G[visited_node][unvisited_node].get(weight, 1.0)
                            if edge_weight < min_distance:
                                min_distance = edge_weight
                                best_next = unvisited_node
                                best_path_node = visited_node
                        except (KeyError, TypeError):
                            if 1.0 < min_distance:
                                min_distance = 1.0
                                best_next = unvisited_node
                                best_path_node = visited_node
            
            if best_next is not None:
                current = best_path_node
                visited.add(best_next)
                path.append(best_next)
                current = best_next
            else:
                break
    
    return path


def dijkstra_order_traversal(G, source, weight: str = 'weight', **kwargs) -> List[int]:
    """
    Visit all nodes in order of their distance from source (Dijkstra-style).
    
    Visits nodes in order of increasing shortest path distance from the source,
    guaranteeing that all reachable nodes are visited.
    
    Parameters
    ----------
    G : networkx.Graph
        The graph to traverse
    source : node
        Starting node for traversal
    weight : str, optional
        Edge attribute to use for weights (default: 'weight')
    **kwargs
        Additional parameters (ignored)
        
    Returns
    -------
    List[int]
        Nodes visited in order of distance from source
    """
    if source not in G:
        raise ValueError(f"Source node {source} not in graph")
    
    import heapq
    
    distances = {node: float('inf') for node in G.nodes()}
    distances[source] = 0
    visited = set()
    path = []
    heap = [(0, source)]
    
    while heap and len(visited) < G.number_of_nodes():
        current_dist, current_node = heapq.heappop(heap)
        
        if current_node in visited:
            continue
            
        visited.add(current_node)
        path.append(current_node)
        
        for neighbor in G.neighbors(current_node):
            if neighbor not in visited:
                try:
                    edge_weight = G[current_node][neighbor].get(weight, 1.0)
                    new_distance = current_dist + edge_weight
                    
                    if new_distance < distances[neighbor]:
                        distances[neighbor] = new_distance
                        heapq.heappush(heap, (new_distance, neighbor))
                except (KeyError, TypeError):
                    new_distance = current_dist + 1.0
                    if new_distance < distances[neighbor]:
                        distances[neighbor] = new_distance
                        heapq.heappush(heap, (new_distance, neighbor))
    
    return path


def weighted_dfs_traversal(G, source, weight: str = 'weight', **kwargs) -> List[int]:
    """
    Depth-first traversal that prioritizes edges with lower weights.
    
    At each node, explores neighbors in order of increasing edge weight,
    ensuring all reachable nodes are visited.
    
    Parameters
    ----------
    G : networkx.Graph
        The graph to traverse
    source : node
        Starting node for traversal
    weight : str, optional
        Edge attribute to use for weights (default: 'weight')
    **kwargs
        Additional parameters (ignored)
        
    Returns
    -------
    List[int]
        Nodes visited in weighted DFS order
    """
    if source not in G:
        raise ValueError(f"Source node {source} not in graph")
    
    visited = set()
    path = []
    
    def weighted_dfs_visit(node):
        if node in visited:
            return
            
        visited.add(node)
        path.append(node)
        
        neighbors_with_weights = []
        for neighbor in G.neighbors(node):
            if neighbor not in visited:
                try:
                    edge_weight = G[node][neighbor].get(weight, 1.0)
                    neighbors_with_weights.append((edge_weight, neighbor))
                except (KeyError, TypeError):
                    neighbors_with_weights.append((1.0, neighbor))
        
        neighbors_with_weights.sort(key=lambda x: x[0])
        
        for _, neighbor in neighbors_with_weights:
            weighted_dfs_visit(neighbor)
    
    weighted_dfs_visit(source)
    
    unvisited = set(G.nodes()) - visited
    for remaining_node in unvisited:
        weighted_dfs_visit(remaining_node)
    
    return path


__all__ = [
    'factor_children',
    'refactor_children',
    'get_signs',
    'get_abs',
    'rotate_children',
    'print_subdivisions',
]

def factor_children(subdivs:tuple) -> tuple:
    def _factor(subdivs, acc):
        for element in subdivs:
            if isinstance(element, tuple):
                _factor(element, acc)
            else:
                acc.append(element)
        return acc
    return tuple(_factor(subdivs, []))

def refactor_children(subdivs:tuple, factors:tuple) -> tuple:
    def _refactor(subdivs, index):
        result = []
        for element in subdivs:
            if isinstance(element, tuple):
                nested_result, index = _refactor(element, index)
                result.append(nested_result)
            else:
                result.append(factors[index])
                index += 1
        return tuple(result), index
    return _refactor(subdivs, 0)[0]

def get_signs(subdivs):
        signs = []
        for element in subdivs:
            if isinstance(element, tuple):
                signs.extend(get_signs(element))
            else:
                signs.append(1 if element >= 0 else -1)
        return signs
    
def get_abs(subdivs):
        result = []
        for element in subdivs:
            if isinstance(element, tuple):
                result.extend(get_abs(element))
            else:
                result.append(abs(element))
        return result

def rotate_children(subdivs: tuple, n: int = 1, preserve_signs: bool = False) -> tuple:
    """Rotates the children of a nested tuple structure.
    
    Args:
        subdivs: Nested tuple structure to rotate
        n: Number of positions to rotate
        preserve_signs: If True, preserves the signs of numbers while rotating their absolute values
    """
    if not preserve_signs:
        factors = factor_children(subdivs)
        n = n % len(factors)
        factors = factors[n:] + factors[:n]
        return refactor_children(subdivs, factors)
    
    signs = get_signs(subdivs)
    abs_values = get_abs(subdivs)
    
    n = n % len(abs_values)
    rotated_values = abs_values[n:] + abs_values[:n]
    
    signed_values = [val * sign for val, sign in zip(rotated_values, signs)]
    
    return refactor_children(subdivs, tuple(signed_values))

def print_subdivisions(subdivs):
    """Format nested tuple structure removing commas."""
    if isinstance(subdivs, (tuple, list)):
        inner = ' '.join(str(print_subdivisions(x)) for x in subdivs)
        return f"({inner})"
    return str(subdivs) 